'''
Tests for adding intention variables
'''

# System imports
import pytest

# 3rd party imports

# local imports
from tests import N
from CoBaIR.bayes_net import BayesNet, load_config

# end file header
__author__ = 'Adrian Lubitz'



def test_context_intention_from_empty():
    """
    Test adding a context and following an intention to an empty config
    """
    bn = BayesNet()
    # I assume this will throw an Error!
    with pytest.raises(AssertionError):
        bn.add_context('the context', {
            'inst_a': 0.3, 'inst_b': 0.3, 'inst_c': 0.4})
    bn.add_intention('some intention')

def test_n_context_intention_from_empty(n=N):
    """
    Test adding n contexts and following an intention to an empty config
    """
    context = 'context'
    instantiations = {
        'inst_a': 0.3, 'inst_b': 0.3, 'inst_c': 0.4}
    bn = BayesNet()
    for i in range(n):
        # I assume this will throw an Error!
        with pytest.raises(AssertionError):
            bn.add_context(f'{context}_{i}', instantiations)
    bn.add_intention('some intention')

def test_intention_context_from_empty():
    """
    Test adding an intention and following a context to an empty config
    """
    bn = BayesNet()
    # I assume this will throw an Error!
    with pytest.raises(AssertionError):
        bn.add_intention('some intention')
    bn.add_context('the context', {
        'inst_a': 0.3, 'inst_b': 0.3, 'inst_c': 0.4})

def test_n_intention_context_from_empty(n=N):
    """
    Test adding n intentions and following a context to an empty config
    """
    intention = 'intention'

    bn = BayesNet()
    for i in range(n):
        # I assume this will throw an Error!
        with pytest.raises(AssertionError):
            bn.add_intention(f'{intention}_{i}')
    bn.add_context('the context', {
        'inst_a': 0.3, 'inst_b': 0.3, 'inst_c': 0.4})

def test_load_from_existing():
    """
    Test if loading a new config overwrites the old
    """
    bn = BayesNet()
    # I assume this will throw an Error!
    with pytest.raises(AssertionError):
        bn.add_context('the context', {
            'inst_a': 0.3, 'inst_b': 0.3, 'inst_c': 0.4})
    bn.add_intention('some intention')
    bn.load('small_example.yml')
    # Assert that it overwrites
    assert bn.config == load_config('small_example.yml')

def test_load_and_save():
    """
    Testing if loaded and saved configs are equal.
    """
    bn = BayesNet()
    bn.load('small_example.yml')
    bn.save('small_example_copy.yml')
    assert load_config('small_example.yml') == load_config('small_example_copy.yml')





"""
This module contains some default functions to discretize incoming evidence

!!! note
    If you want to bindb a discretizer with different default values using 
    `bayes_net.bind_discretization_function` you have to wrap it in a lambda
    function like this: 

    ```
    net.bind_discretization_function('some context', lambda x: binary_decision(x, decision_boundary=0.7))
    ```
"""


def binary_decision(prob_value: float, decision_boundary: float = 0.5, flip: bool = False):
    """
    This function discretizes a probability value into the discrete values True and False.

    Args:
        prob_value: A probability value between 0 and 1
        decision_boundary: Values >= this decision boundary will be discretized as True
        flip: A flag which decides if the discrete result is flipped
    Returns:
        bool:
            A boolean discretization of the given probaility value
    Raises:
        ValueError: A ValueError is raised if the probability value is not in the range [0.0, 1.0]
    """
    if prob_value > 1.0 or prob_value < 0.0:
        raise ValueError(
            f'Probability values have to be in the range of [0.0, 1.0]. Given value is {prob_value}')
    if prob_value >= decision_boundary:
        return True != flip
    else:
        return False != flip



'''
Tests for adding context variables
'''

# System imports
import pytest
from collections import defaultdict

# 3rd party imports

# local imports
from tests import N
from CoBaIR.bayes_net import BayesNet, config_to_default_dict, default_to_regular, load_config

# end file header
__author__ = 'Adrian Lubitz'

# TODO: add more negative cases


def test_add_to_empty(config=None, counter=0):
    """
    Test adding a context to an empty config
    """
    # bn = None  # is needed because bn will only exist in with statement otherwise
    if config == None:
        config = {'intentions': defaultdict(lambda: defaultdict(
            lambda: defaultdict(int))), 'contexts': defaultdict(lambda: defaultdict(float))}
        bn = BayesNet()
    else:
        # with pytest.raises(AssertionError):
        bn = BayesNet(config, validate=False)
    context = f'the context_{counter}'
    instantiations = {
        f'inst_a_{counter}': 0.3, f'inst_b_{counter}': 0.3, f'inst_c_{counter}': 0.4}

    config['contexts'][context] = instantiations

    # I assume this will throw an Error!
    with pytest.raises(AssertionError):
        bn.add_context(context, instantiations)
    # Making sure tmp_config will be maintained
    assert config_to_default_dict(bn.config) == config_to_default_dict(config)


def test_add_n_to_empty(n=N):
    """
    Test adding n contexts to an empty config
    """
    config = {'intentions': defaultdict(lambda: defaultdict(
        lambda: defaultdict(int))), 'contexts': defaultdict(lambda: defaultdict(float))}
    for i in range(n):
        test_add_to_empty(config, i)


@pytest.mark.timeout(1)
def test_add_to_existing(config=None, counter=0):
    """
    Test adding a context to an existing config
    """
    if config == None:
        config = load_config('small_example.yml')
    bn = BayesNet(config)
    context = f'the context_{counter}'
    instantiations = {
        f'inst_a_{counter}': 0.3, f'inst_b_{counter}': 0.3, f'inst_c_{counter}': 0.4}

    instantiation_influences = {
        f'inst_a_{counter}': 0, f'inst_b_{counter}': 0, f'inst_c_{counter}': 0}
    bn.add_context(context, instantiations)
    config['contexts'][context] = instantiations
    for intention in config['intentions']:
        config['intentions'][intention][context] = instantiation_influences
    assert default_to_regular(bn.config) == default_to_regular(config)


# @pytest.mark.timeout(N)
# def test_add_n_to_existing(n=N):
#     """
#     Test adding n contexts to an existing config
#     """
#     config = load_config('small_example.yml')
#     for i in range(n):
#         print('START')
#         test_add_to_existing(config=config, counter=i)
#         print(f'{config=}')

@pytest.mark.timeout(N)
def test_add_to_existing_loaded(config=None, counter=0):
    """
    Test adding a context to an existing config - with empty init and load
    """
    if config == None:
        config = load_config('small_example.yml')
    bn = BayesNet()
    bn.load('small_example.yml')

    context = f'the context_{counter}'
    instantiations = {
        f'inst_a_{counter}': 0.3, f'inst_b_{counter}': 0.3, f'inst_c_{counter}': 0.4}

    instantiation_influences = {
        f'inst_a_{counter}': 0, f'inst_b_{counter}': 0, f'inst_c_{counter}': 0}

    bn.add_context(context, instantiations)
    config['contexts'][context] = instantiations
    for intention in config['intentions']:
        config['intentions'][intention][context] = instantiation_influences
    assert default_to_regular(bn.config) == default_to_regular(config)



'''
Tests for deleting intention variables
'''

# System imports
from copy import deepcopy
import pytest

from collections import defaultdict
# 3rd party imports

# local imports
from tests import N
from CoBaIR.bayes_net import BayesNet, config_to_default_dict, default_to_regular, load_config

# end file header
__author__ = 'Arunima Gopikrishnan'


def test_delete_intention_from_empty():
    """
    Test deleting on a non existing config
    """
    bn = BayesNet()
    intention = "hand over tool"
    # I assume this will throw an Error!
    with pytest.raises(ValueError):
        bn.del_intention(intention)
    assert bn.config == config_to_default_dict()


def test_delete_intention_from_existing_intention():
    """
    Test deleting on an existing config - deleting the name of the intention
    """
    bn = BayesNet()
    bn.load('small_example.yml')
    old_config = deepcopy(bn.config)
    # cnt =0
    new_intention = list(bn.config['intentions'].keys())
    for intention in new_intention:
        if len(bn.config['intentions'].keys()) > 1:
            bn.del_intention(intention)
            assert intention not in bn.config['intentions']
        else:
            # while deleting the last context raises an Exception because there is no context
            with pytest.raises(AssertionError):
                bn.del_intention(intention)
                assert intention not in bn.config['intentions']
    assert bn.config != old_config


def test_delete_to_config_before_adding():
    """
    Test deleting a newly added intention to an existing config
    """
    bn = BayesNet()
    bn.load('small_example.yml')
    old_config = deepcopy(bn.config)
    new_intention = 'add new intention'
    bn.add_intention(new_intention)
    assert new_intention in bn.config['intentions']
    bn.del_intention(new_intention)
    assert new_intention not in bn.config['intentions']


def test_delete_config_before_edit():
    """
    Test deleting on an existing config - only changing the name of the intention
    """
    bn = BayesNet()
    bn.load('small_example.yml')
    new_name = 'entregar la herramienta'
    old_intention = 'hand over tool'
    bn.edit_intention(old_intention, new_name)
    assert new_name in bn.config['intentions']
    assert old_intention not in bn.config['intentions']
    bn.del_intention(new_name)
    assert new_name not in bn.config['intentions']



'''
Tests for adding intention variables
'''

# System imports
from copy import deepcopy
import pytest

# 3rd party imports

# local imports
from tests import N
from CoBaIR.bayes_net import BayesNet, default_to_regular, load_config

# end file header
__author__ = 'Arunima Gopikrishnan'

def test_edit_intention_from_empty():
    """
    Test editing on a non existing config
    """
    bn = BayesNet()
    old_config1 = deepcopy(bn.config)
    intention = 'hand over tool'
    new_name = 'entregar la herramienta'
    # I assume this will throw an Error!
    with pytest.raises(ValueError):
        bn.edit_intention(intention, new_name=new_name)
    assert old_config1 == bn.config

def test_edit_intention_from_existing_new_name():
    """
    Test editing on an existing config - only changing the name of the intention
    """
    bn = BayesNet()
    bn.load('small_example.yml')
    new_name = 'entregar la herramienta'
    old_intention = 'hand over tool'
    old_config = deepcopy(bn.config)
    # I assume this will throw an Error! 
    bn.edit_intention(old_intention, new_name=new_name)
    assert new_name in bn.config['intentions']
    assert old_intention not in bn.config['intentions']






'''
This module provides a class for a two-layer bayes net for context based intention recognition.
'''

# System imports
from __future__ import annotations
import itertools
from collections import defaultdict
from collections.abc import Hashable
from copy import deepcopy
import warnings
import os

# 3rd party imports
import bnlearn as bn
from pgmpy.factors.discrete import TabularCPD
import numpy as np
import yaml
# local imports
from .random_base_count import Counter

# end file header
__author__ = 'Adrian Lubitz'

# https://stackoverflow.com/questions/9169025/how-can-i-add-a-python-tuple-to-a-yaml-file-using-pyyaml


class PrettySafeLoader(yaml.SafeLoader):
    def construct_python_tuple(self, node):
        return tuple(self.construct_sequence(node))


PrettySafeLoader.add_constructor(
    u'tag:yaml.org,2002:python/tuple',
    PrettySafeLoader.construct_python_tuple)


class BayesNet():
    def __init__(self, config: dict = None, bn_verbosity: int = 0, validate: bool = True) -> None:
        '''
        Initializes the BayesNet with the given config.

        Args:
            config: A dict with a config following the config format.
            bn_verbosity: sets the verbose flag for bnlearn. See [bnlearn API](https://erdogant.github.io/bnlearn/pages/html/bnlearn.bnlearn.html?highlight=verbose#bnlearn.bnlearn.make_DAG) for more information
            validate: Flag if the given config should be validated or not. This is necessary to load invalid configs
        '''

        self.valid = False
        self.bn_verbosity = bn_verbosity
        self.discretization_functions = {}

        if config is None:
            validate = False
        config = config_to_default_dict(config)

        # if not config:
        #     self.config = {'intentions': defaultdict(lambda: defaultdict(
        #         lambda: defaultdict(int))), 'contexts': defaultdict(lambda: defaultdict(float))}
        #     return

        self.config = deepcopy(config)
        self.decision_threshold = self.config['decision_threshold']

        if validate:  # this is needed to load invalid configs
            self.validate_config()
        # Translation dicts for context to card number in bnlearn and vice versa
        self._create_value_to_card()
        self._create_card_to_value()
        # Translation dict for the std values to probabilities
        self.value_to_prob = {5: 0.95, 4: 0.75,
                              3: 0.5, 2: 0.25, 1: 0.05, 0: 0.0}

        # initialize the bayes net structure
        self.contexts = self.evidence = list(self.config['contexts'].keys())
        self.intentions = list(self.config['intentions'].keys())
        self.edges = list(itertools.product(self.contexts, self.intentions))
        self._create_evidence_card()

        # create CPTs for the bayes net
        self.cpts = []
        self._create_context_cpts()
        self._create_intention_cpts()
        if self.valid:
            self.DAG = bn.make_DAG(self.edges, CPD=self.cpts,
                                   verbose=self.bn_verbosity)

    def _create_value_to_card(self):
        '''
        Initializes the translation dict for the context values to card numbers for bnlearn
        '''
        self.value_to_card = defaultdict(dict)
        for context, probabilities in self.config['contexts'].items():
            count = 0
            for key, val in probabilities.items():
                self.value_to_card[context][key] = count
                count += 1

    def _create_card_to_value(self):
        '''
        Initializes the backtranslation dict for the context values to card numbers for bnlearn
        '''
        self.card_to_value = defaultdict(dict)
        for context, values in self.value_to_card.items():
            for name, num in values.items():
                self.card_to_value[context][num] = name

    def _create_context_cpts(self):
        '''
        Create the Conditional Probability Tables for all context nodes in the DAG and APPENDS them to self.cpts
        '''
        for context, probabilities in self.config['contexts'].items():
            values = [None] * len(probabilities)
            for value in probabilities:
                values[self.value_to_card[context][value]] = [
                    probabilities[value]]
            self.cpts.append(TabularCPD(variable=context,
                             variable_card=len(probabilities), values=values))

    def _create_intention_cpts(self):
        '''
        Create the Conditional Probability Tables for all intention nodes in the DAG and APPENDS them to self.cpts
        '''
        for intention, context_influence in self.config['intentions'].items():
            values = self._calculate_probability_values(context_influence)
            # create a TabularCPD
            self.cpts.append(TabularCPD(variable=intention, variable_card=2,  # intentions are always binary
                                        # see https://pgmpy.org/factors/discrete.html?highlight=cpd#module-pgmpy.factors.discrete.CPD
                                        values=values,
                                        evidence=self.evidence,
                                        evidence_card=self.evidence_card))

    def _create_evidence_card(self):
        '''
        create the evidence_card for bnlearn
        '''
        self.evidence_card = []
        for evidence_variable in self.evidence:
            self.evidence_card.append(
                len(self.config['contexts'][evidence_variable]))

    def _create_combined_context(self, context_influence: dict) -> dict:
        # TODO: adjust example to use tuples
        """
        Creates a dict with the combined contexts in card index format from context_influence.

        Args:
            context_influence:
                A dict with the influence values for contexts.
                Example: {'speech commands':
                            {'pickup': 5, 'handover': 0, 'other': 0},
                          'human holding object':
                            {True: 1, False: 4},
                          'human activity':
                            {'idle': 4, 'working': 3}
                          }

        Returns:
            dict: A dict with the combined contexts
            Example: {(0, 2): {('pickup', 'working'): 5}), (0, 1): {('pickup', True): 5})}

        """
        combined_context = {}
        for context in context_influence:
            if isinstance(context, tuple):
                combined_context[tuple(map(self.evidence.index, context))
                                 ] = context_influence[context]
        return combined_context

    def _alter_combined_context(self, count: Counter, context_influence: dict, combined_context: dict) -> dict:
        # TODO: adjust example to use tuples
        """
        Overwrites the influence values for the cases of combined influence.

        Args:
            count: A counter that indicates for which combination of context the average is calculated
            context_influence: 
                A dict with the influence values for contexts.
                Example: {'speech commands':
                            {'pickup': 5, 'handover': 0, 'other': 0},
                          'human holding object':
                            {True: 1, False: 4},
                          'human activity':
                            {'idle': 4, 'working': 3}
                          }
            combined_context:
                A dict with the combined contexts
                Example: {(0, 2): {('pickup', 'working'): 5}), (0, 1): {('pickup', True): 5})}

        Returns:
            dict:
                A dict with the adjusted influence values for contexts.
                Example: {'speech commands':
                            {'pickup': 5, 'handover': 0, 'other': 0},
                          'human holding object':
                            {True: 5, False: 4},
                          'human activity':
                            {'idle': 4, 'working': 3}
                          }

        """
        active_case = list(
            map(lambda tup: self.card_to_value[self.evidence[tup[0]]][tup[1]], enumerate(count)))

        altered_context_influence = deepcopy(context_influence)

        for context_tuple, values in combined_context.items():
            combined_case = True
            # There should always be only one key
            value_tuple = list(values.keys())[0]
            for i, context_index in enumerate(context_tuple):
                if active_case[context_index] != value_tuple[i]:
                    combined_case = False
                    break
            if combined_case:
                for i, index in enumerate(context_tuple):
                    altered_context_influence[self.evidence[index]][value_tuple[i]
                                                                    ] = combined_context[context_tuple][value_tuple]
                break
        return altered_context_influence

    def _calculate_probability_values(self, context_influence: dict) -> list:
        # TODO: adjust example to use tuples
        '''
        Calculates the probability values with the given context_influence from the config.

        Influence on the positive case(intention is true) is calculated as the average over all influences for the given context.
        The influence mapping is given in
        self.value_to_prob = {5: 0.95, 4: 0.75,
            3: 0.5, 2: 0.25, 1: 0.05, 0: 0.0}
        Args:
            context_influence:
                A dict with the influence values for contexts.
                Example: {'speech commands':
                            {'pickup': 5, 'handover': 0, 'other': 0},
                          'human holding object':
                            {True: 1, False: 4},
                          'human activity':
                            {'idle': 4, 'working': 3}
                          }
        Returns:
            list:
            A list of lists containing the probability values for the negative and positive respectively.
            Example:

            [[0.416, 0.5, 0.183, 0.266, 0.733, 0.816, 0.5, 0.583, 0.733, 0.816, 0.5, 0.583],

             [0.583, 0.5, 0.816, 0.733, 0.266, 0.183, 0.5, 0.416, 0.266, 0.183, 0.5, 0.416]]
        '''
        # For every intention calculate the average of their influencing contexts
        pos_values = []
        combined_context = self._create_combined_context(context_influence)

        for count in Counter(self.evidence_card):
            # Here I need to average over all the values that are in the config at position count
            average = 0

            # alternate context_influence
            altered_context_influence = self._alter_combined_context(
                count, context_influence, combined_context)

            ####

            for i in range(len(self.evidence_card)):
                average += self.value_to_prob[altered_context_influence[self.evidence[i]
                                                                        ][self.card_to_value[self.evidence[i]][count[i]]]]
            average /= len(self.evidence)
            pos_values.append(average)
        # create neg_values
        neg_values = [1-value for value in pos_values]
        return [neg_values, pos_values]

    def valid_evidence(self, context: str, instantiation) -> tuple[bool, str]:
        """
        Tests if evidence is a valid instantiation for the context.

        Returns a bool if evidence is valid or not and a string with a error message if not valid.
        Args:
            context: a context
            instantiation: an instantiation of the context
        Returns:
            tuple[bool, str]:
            A tuple of bool to indicate validity and str for error message
        """
        if context not in self.config['contexts'] or instantiation is None:
            # ignoring unrelated contexts and Nonetype # TODO: if I ignore it anyways then I can as well say it is valid
            return False, 'ignore'
        if not isinstance(instantiation, Hashable) or not instantiation in self.config['contexts'][context].keys():
            return False, f'{instantiation} is not a valid instantiation for {context}. Must be one of {list(self.config["contexts"][context].keys())}'
        else:
            return True, ''

    def bind_discretization_function(self, context, discretization_function):
        """
        binds a discretization_function to a specific context.


        Args:
            context: One of the possible contexts from the config
            discretization_function: A discretization function which has to take one parameter and return one of the possible discrete context instantiations.
        """
        if context not in self.contexts:
            raise ValueError(
                f'Cannot bind discretization function to {context}. Context does not exist!')
        self.discretization_functions[context] = discretization_function

    def infer(self, evidence, normalized=True, decision_threshold=None) -> dict:
        '''
        infers the probabilities for the intentions with given evidence.

        Args:
            evidence:
                Evidence to infer the probabilities of all intentions.
                Evidence can contain context which is not in the config as well as it must not contain all possible contexts.
                Example:
                    {'speech commands': 'pickup',
                     'human holding object': True,
                     'human activity': 'idle'}
            decision_threshold: a threshold for picking the most likely intention. Must be between 0 and 1. If not given the decision_threshold defined on initialization is taken. 
            normalized: Flag if the returned inference is normalized to sum up to 1.
        Returns:
            tuple:
            Returns the highest ranking intention (or None if decision_threshold is not reached) and a dictionary of intentions and the corresponding probabilities.
        '''
        # check if evidence values are in instantiations and create a card form of bnlearn
        if decision_threshold is None:
            decision_threshold = self.config['decision_threshold']
        card_evidence = {}
        for context, instantiation in evidence.items():
            valid, err_msg = self.valid_evidence(context, instantiation)
            if valid:
                card_evidence[context] = self.value_to_card[context][instantiation]
            elif context in self.discretization_functions and instantiation is not None:
                discrete_instantiation = self.discretization_functions[context](
                    instantiation)
                valid, err_msg = self.valid_evidence(
                    context, discrete_instantiation)
                if valid:
                    card_evidence[context] = self.value_to_card[context][discrete_instantiation]
                else:
                    if not err_msg == 'ignore':  # A discretizer function should still be able to output None
                        raise ValueError(err_msg)
            else:
                if not err_msg == 'ignore':
                    raise ValueError(err_msg)

        if self.valid:
            inference = {}
            for intention in self.intentions:
                # only True values of binary intentions will be saved
                inference[intention] = bn.inference.fit(
                    self.DAG, variables=[intention], evidence=card_evidence, verbose=self.bn_verbosity).values[1]
            if normalized:
                inference = self.normalize_inference(inference)
            max_intention = max(inference, key=inference.get)
            max_intention = max_intention if inference[max_intention] > decision_threshold else None
            return max_intention, inference
        else:
            raise Exception('Configuration is invalid')

    def normalize_inference(self, inference: dict) -> dict:
        '''
        Normalizes the inference to a proper probability distribution.

        Inference which is not normalized will just be normalized for one intention being True or False,
        which leads to uninterpretable results for inference of multiple intentions.

        Args:
            inference: dictionary of intentions and the corresponding probabilities
        Returns:
            dict: dictionary of intentions and the corresponding normalized probabilities.

        '''
        normalized_inference = {}
        probability_sum = sum(inference.values())
        for intention, probability in inference.items():
            normalized_inference[intention] = probability / probability_sum
        return normalized_inference

    def validate_config(self):
        '''
        validate that the current config follows the correct format.

        Raises:
            AssertionError: An AssertionError is raised if the config is not valid.
        '''
        # TODO: add validation that decision_threshold is a float
        # TODO: add validation that apriorio values are float
        # contexts and intentions need to be defined
        assert 'contexts' in self.config, 'Field "contexts" must be defined in the config'
        assert 'intentions' in self.config, 'Field "intentions" must be defined in the config'
        assert len(self.config['contexts']), 'No contexts defined'
        assert len(self.config['intentions']), 'No intentions defined'
        assert isinstance(self.config['decision_threshold'], float) and self.config['decision_threshold'] >= 0 and self.config[
            'decision_threshold'] < 1, 'Decision threshold must be a number between 0 and 1'

        # Intentions need to have influence values for all contexts and their possible instantiations
        for intention, context_influences in self.config['intentions'].items():
            for context, influences in context_influences.items():

                if isinstance(context, str):
                    assert context in self.config[
                        'contexts'], f'Context influence {context} cannot be found in the defined contexts!'
                # assert influences.keys() == self.config['contexts'][context].keys(
                # ), f'An influence needs to be defined for all instantiations! {intention}.{context} does not fit the defined instantiations for {context}'
                for instantiation, influence in influences.items():
                    if type(instantiation) is not tuple:
                        assert 5 >= influence >= 0 and isinstance(
                            influence, int), f'Influence Value for {intention}.{context}.{instantiation} must be an integer between 0 and 5! Is {influence}'
                        assert instantiation in self.config['contexts'][context].keys(
                        ), f'An influence needs to be defined for all instantiations! {intention}.{context}.{instantiation} does not fit the defined instantiations for {context}'
        # Probabilities need to sum up to 1
        for context, instantiations in self.config['contexts'].items():
            for instantiation, value in instantiations.items():
                assert isinstance(
                    value, float), f'Apriori probability of context "{context}.{instantiation}" is not a number'
            assert sum(instantiations.values(
            )) == 1.0, f'The sum of probabilities for context instantiations must be 1 - For "{context}" it is {sum(instantiations.values())}!'
            # This is the config of the currently running BayesNet
        self.valid_config = deepcopy(self.config)
        self.valid = True

    def _create_zero_influence_dict(self, context_with_instantiations: dict) -> defaultdict:
        """
        This uses the context dict from config['contexts'] to instantiate a dict that can be used in config['intentions']['some_context']

        Args:
            context_with_instantiations: a dict holding contexts, their instantiations and corresponding apriori probabilities
        Returns:
            defaultdict:
            A dictionary with zero-initialized influence values for every given context.
            Example:
            {some_context:{
                inst_1:0,
                inst_2:0,
                ...
                inst_3:0
                }
            }
        """
        zeros = defaultdict(lambda: defaultdict(dict))
        for context, instantiations in context_with_instantiations.items():
            for instantiation, value in instantiations.items():
                zeros[context][instantiation] = 0
        return zeros

    def add_context(self, context: str, instantiations: dict):
        """
        This will add a new context to the config and updates the bayesNet.

        Args:
            context: a new context for the config
            instantiations:
                a dict of the instantiations and their corresponding apriori probabilities
                Example:
                    {True: 0.6, False:0.4}
        Raises:
            ValueError: Raises a ValueError if the context already exists in the config
            AssertionError: An AssertionError is raised if the resulting config is not valid.
        """
        # check if context exists already
        if context in self.config['contexts']:
            raise ValueError(
                'Cannot add existing context - use edit_context to edit an existing context')
        # fill in the new context
        self.config['contexts'][context] = instantiations
        # add this context in every intention with instantiations and values beeing zero.
        # for intention in self.config['intentions']:
        #     self.config['intentions'][intention] = {**self.config['intentions'][intention], **self._create_zero_influence_dict(
        #         {context: instantiations})}
        self._transport_context_into_intentions()
        # reinizialize
        self.__init__(self.config)

    def add_intention(self, intention: str):
        """
        This will add a new intention to the config and updates the bayesNet.

        Args:
            intention: Name of a new intention

        Raises:
            ValueError: Raises a ValueError if the intention already exists in the config
            AssertionError: An AssertionError is raised if the resulting config is not valid.
        """
        # check if intention exists already
        if intention in self.config['intentions']:
            raise ValueError(
                'Cannot add existing intention - use edit_intention to edit an existing intention')
        # add in the intention filled with zeros for all contexts
        self.config['intentions'][intention] = defaultdict(
            lambda: defaultdict(int))
        self._transport_context_into_intentions()
        # for context, instantiations_with_values in self.config['contexts'].items():
        #     zeros = self._create_zero_influence_dict(
        #         {context: instantiations_with_values})
        #     self.config['intentions'][intention][context] = zeros[context]
        # reinizialize
        self.__init__(self.config)

    def edit_context(self, context: str, instantiations: dict, new_name: str = None):
        """
        Edits an existing context - this can also be used to remove instantiations

        !!! note
            Changing the name of an instantiation will always set the influence value of this instantiation to zero for all intentions!

        Args:
            context: Name of the context to edit
            instantiations:
                A Dict of instantiations and their corresponding apriori probabilities.
                Example: {True: 0.6, False: 0.4}
            new_name: A new name for the context


        Raises:
            ValueError: Raises a ValueError if the context does not exists in the config
            AssertionError: An AssertionError is raised if the resulting config is not valid.
        """
        # check if context exists already - only then I can edit
        if context not in self.config['contexts']:
            raise ValueError(
                'Cannot edit non existing context - use add_context to add a new context')
        if new_name:  # del old names context
            del (self.config['contexts'][context])
            # rename all occurences in intentions
            for intention in self.config['intentions']:
                old_instantiations = deepcopy(
                    self.config['intentions'][intention][context])
                del (self.config['intentions'][intention][context])
                self.config['intentions'][intention][new_name] = old_instantiations
            context = new_name

        self.config['contexts'][context] = instantiations
        self._remove_context_from_intentions()
        self._transport_context_into_intentions()
        # reinizialize
        self.__init__(self.config)

    def edit_intention(self, intention: str, new_name: str):
        """
        Edits an existing intention.

        Args:
            intention: Name of the intention to edit
            new_name: A new name for the intention

        Raises:
            ValueError: Raises a ValueError if the intention does not exists in the config
            AssertionError: An AssertionError is raised if the resulting config is not valid.
        """
        # check if context exists already - only then I can edit
        if intention not in self.config['intentions']:
            raise ValueError(
                'Cannot edit non existing intention - use add_intention to add a new intention')
        if new_name in self.config['intentions']:
            raise ValueError(
                f'{new_name} exists - cannot be given as the new name for {intention}')
        old_values = deepcopy(self.config['intentions'][intention])
        del (self.config['intentions'][intention])
        self.config['intentions'][new_name] = old_values
        # reinizialize
        self.__init__(self.config)

    def del_context(self, context: str):
        """
        removes a context.

        Args:
            context: Name of the context to delete

        Raises:
            AssertionError: An AssertionError is raised if the resulting config is not valid.
            ValueError: An ValueError is raised if the context is not in self.config.
        """
        # check if context exists already - only then I can edit
        if context not in self.config['contexts']:
            raise ValueError(
                'Cannot delete non existing context - use add_context to add a new context')
        del (self.config['contexts'][context])
        self._remove_context_from_intentions()
        self._transport_context_into_intentions()
        # reinizialize
        self.__init__(self.config)

    def del_intention(self, intention):
        """
        remove an intention.

        Args:
            intention: Name of the intention to delete

        Raises:
            AssertionError: An AssertionError is raised if the resulting config is not valid.
            ValueError: An ValueError is raised if the intention is not in self.config.
        """
        if intention not in self.config['intentions']:
            raise ValueError(
                'Cannot delete non existing intention - use add_intention to add a new intention')
        del (self.config['intentions'][intention])
        # reinizialize
        self.__init__(self.config)

    def save(self, path: str, save_invalid: bool = False):
        """
        saves the config of the bayesNet to a yml file.

        Args:
            path: path to the file the config will be saved in
            save_invalid: Flag to decide if invalid configs can be saved
        Raises:
            ValueError: A ValueError is raised if `save_invalid` is `False` and the config is not valid
        """
        if not self.valid and not save_invalid:
            raise ValueError(
                "saving invalid config is only possible if save_invalid is set to True")

        with open(path, 'w') as save_file:
            yaml.dump(default_to_regular(self.config), save_file)

    def load(self, path: str):
        """
        Loads a config from file and reinitializes the bayesNet.

        Args:
            path: path to the file the config is saved in
        Raises:
            AssertionError: An AssertionError is raised if the resulting config is not valid.
        """
        config = load_config(path)
        # reinizialize with config
        self.__init__(config)

    def change_context_apriori_value(self, context: str, instantiation, value: float):
        """
        Changes the apriori_value for a context instantiation.

        Args:
            context: Name of the context
            instantiation: The instantiation for which the apriori value needs to be changed
            value: the new apriori value
        Raises:
            ValueError: Raises a ValueError if the instantiation does not exists in the config
            AssertionError: An AssertionError is raised if the resulting config is not valid.
        """
        # check if this value already exists because I'm using defaultdict - otherwise you can just add values
        if instantiation in self.config['contexts'][context]:
            self.config['contexts'][context][instantiation] = value
            # reinizialize
            self.__init__(self.config)
        else:
            raise ValueError(
                'change_context_apriori_value can only change values that exist already')

    def change_influence_value(self, intention: str, context: str, instantiation, value: int):
        """
        Change the influence value for a specific instantiation of a context for a specific intention.

        Args:
            intention: Name of the intention
            context: name of the context
            instantiation: The instantiation for which the influence value should be changed
            value: the new influence value. Can be one out of [0, 1, 2, 3, 4, 5]
        Raises:
            ValueError: Raises a ValueError if the instantiation does not exists in the config
            AssertionError: An AssertionError is raised if the resulting config is not valid.
        """
        # check if this value already exists because I'm using defaultdict - otherwise you can just add values
        if instantiation in self.config['intentions'][intention][context]:
            self.config['intentions'][intention][context][instantiation] = value
            self.__init__(self.config)
        else:
            raise ValueError(
                'change_influence_value can only change values that exist already')

    def add_combined_influence(self, intention: str, contexts: tuple, instantiations: tuple, value: int):
        """
        Adds an influence value for a combination of context instantiations.

        Args:
            intention: Name of the intention
            contexts: tuple containing the names of the contexts
            instantiations: tuple of context instantiations for which the influence value should be set
            value: influence value. Can be one out of [0, 1, 2, 3, 4, 5]
        Raises:
            ValueError: Raises a ValueError if the instantiation does not exists in the config
            AssertionError: An AssertionError is raised if the resulting config is not valid.
        """
        for i, instantiation in enumerate(instantiations):
            if instantiation not in self.config['intentions'][intention][contexts[i]]:
                raise ValueError(
                    'add_combined_influence can only combine context instantiations that already exist')
        self.config['intentions'][intention][contexts][instantiations] = value
        self.__init__(self.config)

    def del_combined_influence(self, intention: str, contexts: tuple, instantiations: tuple):
        """
        Adds an influence value for a combination of context instantiations.

        Args:
            intention: Name of the intention
            contexts: tuple containing the names of the contexts
            instantiations: tuple of context instantiations
        Raises:
            ValueError: Raises a ValueError if the instantiation does not exists in the config
            AssertionError: An AssertionError is raised if the resulting config is not valid.
        """
        if instantiations not in self.config['intentions'][intention][contexts]:
            raise ValueError(
                'remove_combined_influence can only remove combined context instantiations that already exist')
        del (self.config['intentions'][intention][contexts])

    def _transport_context_into_intentions(self):
        """
        Transports contexts and their instantiations defined in the config['contexts'] into config['intentions'] as influencing context if not present.
        """
        for context in self.config['contexts']:
            for instantiation in self.config['contexts'][context]:
                for intention in self.config['intentions']:
                    if instantiation not in self.config['intentions'][intention][context]:
                        # This only works if it is a defaultdict
                        self.config['intentions'][intention][context][instantiation] = 0

    def _remove_context_from_intentions(self):
        """
        This removes context or instantiations after removing/changing instantiations and/or context.
        """
        # This is a hack because you can't edit while iterating a dict
        contexts_to_remove_from_intentions = []
        context_instantiations_to_remove_from_intentions = []
        for intention in self.config['intentions']:
            for context in self.config['intentions'][intention]:
                if context not in self.config['contexts']:
                    contexts_to_remove_from_intentions.append(
                        (intention, context))
                else:
                    for instantiation in self.config['intentions'][intention][context]:
                        if instantiation not in self.config['contexts'][context]:
                            context_instantiations_to_remove_from_intentions.append(
                                (intention, context, instantiation))
        for intention, context in contexts_to_remove_from_intentions:
            del (self.config['intentions'][intention][context])
        for intention, context, instantiation in context_instantiations_to_remove_from_intentions:
            del (self.config['intentions'][intention][context][instantiation])

    def change_decision_threshold(self, decision_threshold):
        """
        Changes the decision threshold in the config.
        Args:
            decision_threshold: The new decision threshold.
        Raises:
            AssertionError: An AssertionError is raised if the resulting config is not valid.
        """
        self.config['decision_threshold'] = decision_threshold
        self.__init__(self.config)


def config_to_default_dict(config: dict = None):
    """
    This casts a config given as dict into a defaultdict.

    Args:
        config: A dict with a config following the config format.
    Returns:
        defaultdict:
            a defaultdict containing the config
    """
    if not config:
        config = {}
    new_config = {'intentions': defaultdict(lambda: defaultdict(
        lambda: defaultdict(int))), 'contexts': defaultdict(lambda: defaultdict(float))}
    if 'contexts' in config:
        for context in config['contexts']:
            for instantiation, value in config['contexts'][context].items():
                new_config['contexts'][context][instantiation] = value
    if 'intentions' in config:
        for intention in config['intentions']:
            # HERE: there is the chance that there is no context yet - write intention once
            new_config['intentions'][intention] = defaultdict(
                lambda: defaultdict(int))
            for context in config['intentions'][intention]:
                for instantiation, value in config['intentions'][intention][context].items():
                    new_config['intentions'][intention][context][instantiation] = value
    if 'decision_threshold' in config:
        new_config['decision_threshold'] = config['decision_threshold']
    else:
        new_config['decision_threshold'] = 0.0

    return new_config


def load_config(path):
    """
    Helper function to load a config.

    Args:
        path: path to the file the config is saved in
    Returns:
        defaultdict:
            a defaultdict containing the config
    """

    # if os.path.splitext(path)[-1] != ".yml":
    #     raise TypeError(
    #         'Invalid format file - only supporting yml files')
    with open(path) as stream:
        return config_to_default_dict(yaml.load(stream, Loader=PrettySafeLoader))

# https://stackoverflow.com/questions/26496831/how-to-convert-defaultdict-of-defaultdicts-of-defaultdicts-to-dict-of-dicts-o


def default_to_regular(d):
    """
    This casts a defaultdict to a regular dict which is needed for saving as yml file.

    Args:
        d: the dict which should be casted
    Returns:
        dict:
            a regular dict casted from the defaultdict
    """
    # casts dicts or default dicts because otherwise it will stop at the first dict and if that has another defaultdict in it - that won't cast
    if isinstance(d, defaultdict) or isinstance(d, dict):
        d = {k: default_to_regular(
            v) for k, v in d.items() if not isinstance(v, dict) or v}
    return d



'''
This module helps to increment in a CPT with random number of cards per evidence.
'''

# System imports

# 3rd party imports

# local imports

# end file header
__author__ = 'Adrian Lubitz'


class Counter():
    """A counter for different bases on different positions"""

    def __init__(self, evidence_cards: list) -> None:
        '''
        Takes a list of ints (int must be >=2) and creates the counter

        Args:
            evidence_cards: list of ints as mapping of base to position
        '''
        self.evidence_cards = evidence_cards
        self.current_evidence = [0] * len(self.evidence_cards)
        self.first_call = True

    def __iter__(self):
        """Returns self as a iterator"""
        return self

    def __next__(self, start=-1):
        '''
        increment by one.
        '''
        # Corner Cases: First Call
        if self.first_call:
            self.first_call = False
            return self.current_evidence
        # Corner Cases: Last Call
        if start < -1*len(self.evidence_cards):
            raise StopIteration()
        # Std Case: All other calls
        self.current_evidence[start] += 1
        while(self.current_evidence[start] == self.evidence_cards[start]):
            self.current_evidence[start] = 0
            self.__next__(start-1)
        return self.current_evidence


if __name__ == '__main__':
    # Testing
    for count in Counter([4, 2, 3, 2]):
        print(count)



'''
Tests for deleting context variables
'''

# System imports
from copy import deepcopy
import pytest

from collections import defaultdict
# 3rd party imports

# local imports
from tests import N
from CoBaIR.bayes_net import BayesNet, config_to_default_dict, default_to_regular, load_config

# end file header
__author__ = 'Arunima Gopikrishnan'


def test_delete_context_from_empty():
    """
    Test deleting on a non existing config
    """
    bn = BayesNet()
    context = "speech commands"
    # I assume this will throw an Error!
    with pytest.raises(ValueError):
        bn.del_context(context)
    assert bn.config == config_to_default_dict()


def test_delete_context_from_existing_context():
    """
    Test deleting on an existing config - deleting the name of the context and instantiations
    """
    bn = BayesNet()
    bn.load('small_example.yml')
    old_config = deepcopy(bn.config)
    # cnt =0
    new_contexts = list(bn.config['contexts'].keys())
    for context in new_contexts:
        if len(bn.config['contexts'].keys()) > 1:
            bn.del_context(context)
            assert context not in bn.config['contexts']
        else:
            # while deleting the last context raises an Exception because there is no context
            with pytest.raises(AssertionError):
                bn.del_context(context)
                assert context not in bn.config['contexts']
    assert bn.config != old_config


def test_delete_config_before_adding():
    """
    Test deleting a newly added context to an existing config
    """
    bn = BayesNet()
    bn.load('small_example.yml')
    old_config = deepcopy(bn.config)
    new_context = 'add new context'
    bn.add_context(new_context, {
        'inst_a': 0.2, 'inst_b': 0.3, 'inst_c': 0.5})
    assert new_context in bn.config['contexts']
    bn.del_context(new_context)
    assert new_context not in bn.config['contexts']


def test_delete_to_config_before_edit():
    """
    Test deleting on an existing config - only changing the name of the context
    """
    bn = BayesNet()
    bn.load('small_example.yml')
    new_name = 'commandos de speechos'
    old_context = 'speech commands'
    instantiations = deepcopy(bn.config['contexts'][old_context])
    bn.edit_context(old_context, instantiations, new_name)
    assert new_name in bn.config['contexts']
    assert old_context not in bn.config['contexts']
    bn.del_context(new_name)
    assert new_name not in bn.config['contexts']



'''
Tests for new combined context influence
'''

# System imports
from copy import deepcopy
from matplotlib.style import context
import pytest

from collections import defaultdict

# local imports
from tests import N
from CoBaIR.bayes_net import BayesNet, default_to_regular, load_config

# end file header
__author__ = 'Arunima Gopikrishnan'

config = {'intentions': defaultdict(lambda: defaultdict(
        lambda: defaultdict(int))), 'contexts': defaultdict(lambda: defaultdict(float))}
def test_new_combined_influence_into_empty():
    """
    Test adding combined context influence on a non existing config
    """
    bn = BayesNet()
    intention = 'hand over tool'
    contexts =('speech commands','human activity')
    instantiations =('pickup','working')
    value = 5
    # I assume this will throw an Error!
    with pytest.raises(ValueError):
        bn.add_combined_influence(intention, contexts, instantiations,value)


def test_change_influence_value_in_existing_new_combined_context_influence():
    """
    Test alter existing combined context influence from the loaded config's existing combined context influence
    """
    
    bn = BayesNet()
    bn.load('small_example.yml')
    old_config = deepcopy(bn.config)
 #   for intention, context_influence in bn.config['intentions'].items():
#            bn._create_combined_context(context_influence)
    intention = 'pick up tool'
    contexts = ('speech commands','human activity')
    instantiations = ('pickup','working')
    value = 1
    bn.add_combined_influence(intention, contexts, instantiations,value)
    assert intention in bn.config['intentions']
    bn.add_combined_influence(intention, contexts, instantiations,value)
    assert value not in bn.config
    assert old_config != bn.config


def test_new_combined_context_influence_from_existing_combined_context_influence():
    """
    Test adding combined context influence from the loaded config's existing combined context influence
    """
    
    bn = BayesNet()
    bn.load('small_example.yml')
    for intention, context_influence in bn.config['intentions'].items():
        bn._create_combined_context(context_influence)
    intention = 'hand over tool'
    contexts = ('speech commands','human activity')
    instantiations = ('pickup','idle')
    value=2
    bn.add_combined_influence(intention, contexts, instantiations,value)
    assert context_influence != bn.config
    

def test_create_combined_context():
    """
    Test creating combined context
    """
    
    bn = BayesNet()
    bn.load('small_example.yml')
    combined_context = "{(0, 2): defaultdict(<class 'int'>, {('pickup', 'working'): 5}), (0, 1): defaultdict(<class 'int'>, {('pickup', True): 4})}"
    #combined = json.loads(combined_context)
    #combined_context = "{(0, 1): defaultdict(<class 'int'>, {('pickup', True): 4}), (0, 2): defaultdict(<class 'int'>, {('pickup', 'working'): 5})}"
    for intention, context_influence in bn.config['intentions'].items():
            bn._calculate_probability_values(context_influence)
    data = bn._create_combined_context(context_influence)
    assert combined_context == str(data)


def test_alter_combined_context():
    """
    Test alter existing combined context
    """
        
    bn = BayesNet()
    bn.load('small_example.yml')
    combined_context = "{(0, 2): defaultdict(<class 'int'>, {('pickup', 'working'): 5}), (0, 1): defaultdict(<class 'int'>, {('pickup', True): 4})}"
    altered_combined_context = "{(0, 2): defaultdict(<class 'int'>, {('pickup', 'working'): 2}), (0, 1): defaultdict(<class 'int'>, {('pickup', True): 4})}"
    for intention, context_influence in bn.config['intentions'].items():
            bn._calculate_probability_values(context_influence)
    data = bn._create_combined_context(context_influence)
   # data = bn._calculate_probability_values(defaultdict(<function config_to_default_dict.<locals>.<lambda> at 0x0000017F41ED0A60>, {'speech commands': defaultdict(<class 'int'>, {'handover': 0, 'other': 0, 'pickup': 5}), 'human holding object': defaultdict(<class 'int'>, {False: 4, True: 1}), ('speech commands', 'human activity'): defaultdict(<class 'int'>, {('pickup', 'working'): 5}), 'human activity': defaultdict(<class 'int'>, {'idle': 4, 'working': 3}), ('speech commands', 'human holding object'): defaultdict(<class 'int'>, {('pickup', True): 4})}))
    assert combined_context == str(data)
    bn = BayesNet()
    bn.load('small_example_altered.yml')
    for intention, context_influence in bn.config['intentions'].items():
            bn._calculate_probability_values(context_influence)
    data = bn._create_combined_context(context_influence)
    assert altered_combined_context == str(data)


    
def test_calculate_probability_values():
    """
    Test calculating context probability values
    Influence value to probability = {5: 0.95, 4: 0.75, 3: 0.5, 2: 0.25, 1: 0.05, 0: 0.0}
    """
    
    bn = BayesNet()
    bn.load('small_example.yml')
    for intention, context_influence in bn.config['intentions'].items():
           (pos,neg) = bn._calculate_probability_values(context_influence)
    print(pos,neg)
    bn = BayesNet()
    bn.load('small_example_altered.yml')
    for intention, context_influence in bn.config['intentions'].items():
           (alter_pos,alter_neg) = bn._calculate_probability_values(context_influence)
    print(alter_pos,alter_neg)
    assert alter_pos != pos

    



'''
This module is a GUI configurator to create configurations for context based intention recognition - it can as well be used in a live mode to test the configuration
'''

# System imports
import sys
from collections import defaultdict
import tkinter as tk
from tkinter import filedialog as fd
from tkinter.simpledialog import Dialog
from tkinter import ttk
from copy import deepcopy
from types import FunctionType as function
from pathlib import Path
import itertools

# 3rd party imports
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas
import pyqtgraph as pg
from PyQt5 import QtWidgets, uic
from PyQt5.QtWidgets import *
from PyQt5.QtCore import Qt, QStringListModel
from PyQt5.QtGui import QFont, QFontMetrics
import yaml
import numpy as np


# local imports
from .bayes_net import BayesNet, load_config


# end file header
__author__ = 'Adrian Lubitz'


class NewIntentionDialog(QDialog):
    """Dialog Window for new Intention"""

    def __init__(self, parent=None, intention: str = None) -> None:
        """
        Extends the Constructor of Dialog to use an already existing intention.

        If an intention is given it will be filled in the corresponding text field.

        Args:
            intention: An intention that will be filled in the dialog
        """
        dialog = QDialog()
        dialog.deleteLater()
        self.intention = intention
        super().__init__(parent)
        self.result = None
        self.setModal(True)
        self.body()
        self.show()

    def body(self):
        """
        Sets the Layout.

        Args:
            master: the master window this dialog belongs to
        Returns:
            tk.Entry:
                the initial focus
        """
        uic.loadUi(Path(Path(__file__).parent, 'NewIntention.ui'), self)
        self.intention_entry = self.findChild(QLineEdit, 'lineEdit')
        self.error_label = self.findChild(QLabel, 'label_2')
        self.error_label.setAlignment(Qt.AlignCenter)
        self.error_label.setStyleSheet("color: red")
        if self.intention:
            self.intention_entry.setText(self.intention)
        return self.intention_entry  # initial focus

    def get_result(self):
        """
        Applies the result to hand it over to the master
        """
        self.error_label.setText("")
        result = self.intention_entry.text().strip()
        if not result:
            self.error_label.setText("Intentions cannot be empty")
        else:
            self.accept()
        return result


class NewCombinedContextDialog(QDialog):
    """Dialog Window for new combined Context influence"""

    def __init__(self, parent=None, intentions: dict = None) -> None:
        """
        Extends the Constructor of Dialog to make config available.

        Args:
            intentions:
                The intentions part of the config
                Example:
                    {hand over tool:
                        human activity:
                            idle: 4
                            working: 1
                        speech commands:
                            handover: 5
                            other: 0
                            pickup: 0
                    pick up tool:
                        speech commands:
                            handover: 0
                            other: 0
                            pickup: 5
                        ? !!python/tuple
                        - speech commands
                        - human activity
                            : ? !!python/tuple
                            - pickup
                            - working
                                : 5
                        human activity:
                            idle: 4
                            working: 3
                    }
        """
        self.intentions = deepcopy(intentions)
        self.original_instantiations = defaultdict(dict)
        super().__init__(parent)
        self.result = None
        self.body()
        self.show()

    def body(self):
        """
        Sets the Layout.

        Args:
            master: the master window this dialog belongs to
        """
        uic.loadUi(Path(Path(__file__).parent,
                   'NewCombinedContextInfluence.ui'), self)

        values = list(self.intentions.keys())

        self.intention_selection = self.findChild(QComboBox, 'comboBox')
        self.intention_selection.addItems(values)
        self.intention_selection.setCurrentIndex(0)
        # Value
        self.value_selection = self.findChild(QComboBox, 'comboBox_2')
        # Button
        self.additional_context = self.findChild(QPushButton, 'pushButton')
        self.additional_context.clicked.connect(self.new_instantiation)
        self.error_label = self.findChild(QLabel, 'label_5')
        self.error_label.setAlignment(Qt.AlignCenter)
        self.error_label.setStyleSheet("color: red")
        # contexts
        self.contexts = {}
        self.context_frame = self.findChild(QFrame, 'frame')
        self.grid_layout = QGridLayout()
        self.context_frame.setLayout(self.grid_layout)
        for context, instantiations in list(self.intentions.values())[0].items():
            if not isinstance(context, tuple):
                self.contexts[context] = instantiations
        self.context_selections = []
        self.context_menus = []
        self.instantiation_selections = []
        self.instantiation_menus = []
        for i in range(2):
            self.new_instantiation()

    def context_selected(self, context: str, i: int):
        """
        Callback for context selection in dropdown

        Args:
            context: A context name
            i: the position of the dropdown menu
        """
        self.instantiation_menus[i].setVisible(False)
        self.instantiation_menus[i].deleteLater()

        for inst in self.contexts[context]:
            self.original_instantiations[context][str(inst)] = inst

        instantiations = list(self.original_instantiations[context].keys())
        instantiations = list(map(str, instantiations))

        self.instantiation_selections[i] = QComboBox(self.context_frame)
        self.instantiation_selections[i].clear()
        self.instantiation_selections[i].addItems(instantiations)
        self.instantiation_selections[i].setCurrentText(
            self.instantiation_selections[i].currentText())

        self.grid_layout.addWidget(self.instantiation_selections[i], i+1, 1)

        self.instantiation_selections[i].setCurrentText(
            str(list(self.contexts[context].keys())[0]))

        # available context -> every dropdown should only have available + selected
        available_context = self._eval_available_context()
        for i_m, menu in enumerate(self.context_menus):
            menu.setVisible(False)
            menu.deleteLater()
            contexts = [self.context_selections[i_m].currentText()] + \
                available_context
            self.context_selections[i_m] = QComboBox(self.context_frame)
            self.context_selections[i_m].addItems(contexts)
            self.context_selections[i_m].setCurrentText(
                self.context_selections[i_m].currentText())
            self.context_selections[i_m].currentTextChanged.connect(
                lambda text, i=i_m: self.context_selected(text, i))
            self.grid_layout.addWidget(self.context_selections[i_m], i_m+1, 0)

    def _eval_available_context(self):
        """
        evaluate which contexts are available for the dropdown

        Returns:
            list:
                A list of available contexts
        """
        available_context = list(self.contexts.keys())
        for selection in self.context_selections:
            available_context.remove(selection.currentText())
        return available_context

    def new_instantiation(self):
        """
        Creates new Entries to input more context instantiations
        """
        available_context = self._eval_available_context()
        i = len(self.context_selections)
        if not available_context:
            self.error_label.setText('No more context available')
            return
        self.context_selections.append(QComboBox(self.context_frame))
        self.context_selections[-1].addItems(available_context)
        self.context_selections[-1].setCurrentIndex(0)
        self.context_selections[-1].currentTextChanged.connect(
            lambda text, i=i: self.context_selected(text, i))
        self.context_menus.append(self.context_selections[-1])
        self.grid_layout.addWidget(self.context_menus[-1], i+1, 0)

        instantiations = list(self.contexts[available_context[0]].keys())
        instantiations = list(map(str, instantiations))
        self.instantiation_selections.append(QComboBox(self.context_frame))
        self.instantiation_selections[-1].addItems(instantiations)
        self.instantiation_selections[-1].setCurrentIndex(0)
        self.instantiation_menus.append(self.instantiation_selections[-1])
        self.grid_layout.addWidget(self.instantiation_menus[-1], i+1, 1)

        self.context_selected(self.context_selections[-1].currentText(), i)

    def get_result(self):
        """
        Applies the result to hand it over to the master
        """
        intention = self.intention_selection.currentText()
        value = self.value_selection.currentText()
        contexts = []
        instantiations = []
        for i, context_selection in enumerate(self.context_selections):
            contexts.append(context_selection.currentText())
            instantiations.append(
                self.original_instantiations[context_selection.currentText()][self.instantiation_selections[i].currentText()])
        # intention: str, contexts: tuple, instantiations: tuple, value: int
        result = {'intention': intention, 'value': int(value), 'contexts': tuple(
            contexts), 'instantiations': tuple(instantiations)}
        self.accept()
        return result


class NewContextDialog(QDialog):
    """Dialog Window for new Context"""

    def __init__(self, parent=None, predefined_context: dict = None) -> None:
        """
        Extends the Constructor of Dialog to use already existing context and the corresponding instantiations and values.

        If context and their corresponding instantiations and values are given it will be filled in the corresponding text fields.

        Args:
            predefined_context:
                Context with the corresponding instantiations.
                Example: {'speech commands': {
                    'pickup': 0.2, 'handover': 0.2, 'other': 0.6}}
        """
        dialog = QDialog()
        dialog.deleteLater()
        self.predefined_context = deepcopy(predefined_context)
        super().__init__(parent)
        self.result = None
        self.setModal(True)
        self.body()
        self.show()

    def body(self):
        """
        Sets the Layout.

        Args:
            master: the master window this dialog belongs to
        Returns:
            tk.Entry:
                the initial focus
        """
        uic.loadUi(Path(Path(__file__).parent, 'NewContext.ui'), self)
        self.grid_layout_2 = self.findChild(QGridLayout, 'gridLayout_2')
        self.context_entry = self.findChild(QLineEdit, 'context_entry')
        self.instantiations_frame = self.findChild(
            QFrame, 'instantiations_frame')
        self.error_label = self.findChild(QLabel, 'label_5')
        self.error_label.setAlignment(Qt.AlignCenter)
        self.error_label.setStyleSheet("color: red")
        self.instantiations = []
        self.shown_instantiations = 0
        self.grid_layout = QGridLayout()
        self.instantiations_frame.setLayout(self.grid_layout)
        self.instantiations_frame.setSizePolicy(
            QSizePolicy.Expanding, QSizePolicy.MinimumExpanding)
        # TODO: Fill the entries here if predefined_context is given!
        if self.predefined_context:
            # it's always only one new context
            context = list(self.predefined_context.keys())[0]
            instantiations = self.predefined_context[context]
            self.context_entry.setText(context)
        else:
            instantiations = {}
        while(self.shown_instantiations < 2 or instantiations):

            name_entry = QLineEdit()
            probability_entry = QLineEdit()
            remove_button = QPushButton('-')
            remove_button.clicked.connect(
                lambda _, btn=remove_button: self.remove_instantiation(btn))
            if instantiations:
                # get name
                name = list(instantiations.keys())[0]
                # get value
                value = instantiations[name]
                # set entries
                name_entry.setText(str(name))
                probability_entry.setText(str(value))
                # del entry
                del(instantiations[name])
            self.grid_layout.addWidget(
                name_entry, self.shown_instantiations+1, 0)
            self.grid_layout.addWidget(
                probability_entry, self.shown_instantiations+1, 1)
            self.grid_layout.addWidget(
                remove_button, self.shown_instantiations+1, 2)
            self.instantiations.append(
                (name_entry, probability_entry, remove_button))
            self.shown_instantiations += 1
        self.more = self.findChild(QPushButton, 'pushButton')
        self.more.clicked.connect(self.new_instantiation)
        return self.context_entry

    def remove_instantiation(self, remove_button):
        """
        Callback for the remove_button
        Args:
            remove_button: the remove_button clicked
        """
        # find the index of the row containing the remove_button
        self.error_label.setText("")
        index = -1
        for i, (_, _, btn) in enumerate(self.instantiations):
            if btn is remove_button:
                index = i
                break
        if index != -1:
            # remove the row from the layout and from the instantiations list
            name_entry, probability_entry, _ = self.instantiations.pop(index)
            self.grid_layout.removeWidget(name_entry)
            self.grid_layout.removeWidget(probability_entry)
            self.grid_layout.removeWidget(remove_button)
            name_entry.deleteLater()
            probability_entry.deleteLater()
            remove_button.deleteLater()
            self.shown_instantiations -= 1
        self.grid_layout.update()

    def new_instantiation(self):
        """
        Creates two new Entries to input more instantiations
        """
        name_entry = QLineEdit(self.instantiations_frame)
        probability_entry = QLineEdit(self.instantiations_frame)
        remove_button = QPushButton('-', self.instantiations_frame)
        remove_button.clicked.connect(
            lambda _, btn=remove_button: self.remove_instantiation(btn))
        self.instantiations.append(
            (name_entry, probability_entry, remove_button))
        row_count = self.grid_layout.rowCount()
        self.grid_layout.addWidget(name_entry, row_count, 0)
        self.grid_layout.addWidget(probability_entry, row_count, 1)
        self.grid_layout.addWidget(remove_button, row_count, 2)
        self.shown_instantiations += 1
        self.grid_layout.update()
        # Raise ValueError if both name_entry and probability_entry are empty
        if not name_entry.text() and not probability_entry.text():
            self.error_label.setText(
                "Both name and probability cannot be empty.")
        elif not name_entry.text():
            self.error_label.setText("Name cannot be empty.")
        elif not probability_entry.text():
            self.error_label.setText("Probability cannot be empty.")

    def get_result(self):
        result = defaultdict(lambda: defaultdict(dict))
        errors = []
        context_entry = self.context_entry.text().strip()
        if not context_entry:
            errors.append("Context cannot be empty.")
        else:
            for i, instantiation in enumerate(self.instantiations):
                key = instantiation[0].text().strip()
                value = instantiation[1].text().strip()
                if not key and not value:
                    errors.append(
                        f"Name and probability cannot be empty in row {i+1}.")
                elif not key:
                    errors.append(f"Name cannot be empty in row {i+1}.")
                elif not value:
                    errors.append(f"Probability cannot be empty in row {i+1}.")
                else:
                    try:
                        value = float(value)
                    except ValueError:
                        errors.append(
                            f"Probability must be a number in row {i+1}.")
                    else:
                        result[self.context_entry.text().strip()][key] = value
        if errors:
            self.error_label.setText("\n".join(errors))
        else:
            self.accept()
        return result


class Configurator(QtWidgets.QMainWindow):
    '''
    GUI configurator to create configurations for context based intention recognition.
    It can as well be used in a live mode to test the configuration
    '''

    def __init__(self, config: dict = None, *args, **kwargs):
        '''
        Setting up the GUI

        Args:
            config: A dict with a config following the config format.
        '''
        self.app = QtWidgets.QApplication(sys.argv)
        QtWidgets.QMainWindow.__init__(self, *args, **kwargs)
        # self.figure = plt.figure()
        # self.canvas = FigureCanvas(self.figure)
        # creating a graph item
        # setting configuration options
        pg.setConfigOptions(antialias=True)
        # creating graphics layout widget
        self.win = pg.GraphicsLayoutWidget()
        # adding view box to the graphic layout widget
        self.view = self.win.addViewBox()
        self.graph_item = TwoLayerGraph()
        self.view.addItem(self.graph_item)
        self.setup_layout()
        self.bayesNet = BayesNet(config)
        self.create_fields()
        self.show()  # Show the GUI

    def set_error_label_red(self):
        """
        setting the alignment and color of the error label
        """
        self.error_label.setAlignment(
            Qt.AlignCenter)  # TODO: no changes of style in functional code!
        self.error_label.setStyleSheet("color: red")

    def create_fields(self):
        """
        Creates all necessary fields in the GUI.

        This should be used whenever the config is changed.
        It reads all values from the config and adjusts the GUI accordingly.
        """
        self.set_context_dropdown(self.bayesNet.config['contexts'].keys())

        self.set_influencing_context_dropdown(
            self.bayesNet.config['contexts'].keys())

        self.set_intention_dropdown(self.bayesNet.config['intentions'].keys())
        self.adjust_button_visibility()
        self.set_decision_threshold()
        self.set_context_dropdown(self.bayesNet.config['contexts'].keys())

        self.set_influencing_context_dropdown(
            self.bayesNet.config['contexts'].keys())

        self.set_intention_dropdown(self.bayesNet.config['intentions'].keys())
        self.adjust_button_visibility()
        self.set_decision_threshold()
        self.fill_advanced_table()
        self.draw_graph()

    def set_decision_threshold(self):
        """
        This sets value in the decision threshold entry from the config
        """
        self.decision_threshold_entry.setText(
            str(self.bayesNet.config['decision_threshold']))

    def adjust_button_visibility(self):
        """
        Adjusts if buttons are visible or not.

        Buttons for edit and delete will only be visible if there is a corresponding intention or context already created.
        """
        if self.bayesNet.config['contexts']:
            # set visible
            self.edit_context_button.show()
            self.delete_context_button.show()
            self.grid_layout.addWidget(self.new_context_button, 0, 4)
        else:
            self.edit_context_button.hide()
            self.delete_context_button.hide()
            self.grid_layout.addWidget(self.new_context_button, 0, 2)
        if self.bayesNet.config['intentions']:
            # set visible
            self.edit_intention_button.show()
            self.delete_intention_button.show()
            self.grid_layout.addWidget(self.new_intention_button, 2, 6)
        else:
            self.edit_intention_button.hide()
            self.delete_intention_button.hide()
            self.grid_layout.addWidget(self.new_intention_button, 2, 4)

    def new_context(self):
        """
        Open a new Dialog to create new contexts.
        """
        # open small dialog to create context
        dialog = NewContextDialog(self)

        def update_and_close():
            self.error_label.setText("")
            result = dialog.get_result()
            if not result:
                self.error_label.setText(
                    "At least one instantiation is required.")
                return
            try:
                old_context_name = list(result.keys())[0]
                new_instantiations = result[old_context_name]
                if new_instantiations:
                    self.bayesNet.add_context(
                        old_context_name, new_instantiations)
            except AssertionError as e:
                self.error_label.setText(str(e))
            # update view!
            self.create_fields()
            self.context_selection.setCurrentText(old_context_name)
            # Explicit call is necessary because setCurrentText seems not to trigger the callback
            self.context_selected(old_context_name)
            dialog.accept()

        ok_button = dialog.findChild(QPushButton, "pushButton_2")
        ok_button.setDefault(True)
        ok_button.clicked.connect(update_and_close)
        cancel_button = dialog.findChild(QPushButton, "pushButton_3")
        cancel_button.clicked.connect(dialog.reject)
        dialog.exec_()

    def edit_context(self):
        """
        Edit the currently selected context.

        !!! note
        Changing the name of an instantiation will always set the influence value of this instantiation to zero for all intentions!

        !!! note
        The GUI can only handle strings for now. This means every instantiation name will be casted to a string upon editing.
        """
        # TODO: renaming instantiations should not neccesarily put influence values to zero - check cases
        # TODO: this will always set the instantiations as Strings
        # Open the new Context dialog with prefilled values
        # open small dialog to create context

        context = self.context_selection.currentText()
        instantiations = self.bayesNet.config['contexts'][context]
        dialog = NewContextDialog(self, predefined_context={
            context: instantiations})

        def update_and_close():
            self.error_label.setText("")
            result = dialog.get_result()
            if not result:
                self.error_label.setText(
                    "At least one instantiation is required.")
                return
            try:
                old_context_name = list(result.keys())[0]
                new_instantiations = result[old_context_name]
                if new_instantiations:
                    self.bayesNet.edit_context(
                        context, new_instantiations, old_context_name)
            except (ValueError, AssertionError) as e:
                self.error_label.setText(str(e))
            self.create_fields()
            self.context_selection.setCurrentText(old_context_name)
            self.context_selected(old_context_name)
            dialog.accept()
        ok_button = dialog.findChild(QPushButton, "pushButton_2")
        ok_button.setDefault(True)
        ok_button.clicked.connect(update_and_close)
        cancel_button = dialog.findChild(QPushButton, "pushButton_3")
        cancel_button.clicked.connect(dialog.reject)
        dialog.exec_()

    def delete_context(self):
        """"
        Deletes the currently selected context.
        """
        self.error_label.setText("")
        context = self.context_dropdown.currentText()
        try:
            self.bayesNet.del_context(context)
        except AssertionError as e:
            self.error_label.setText(str(e))
        self.create_fields()

    def new_intention(self):
        """
        Open a new Dialog to create new intentions.
        """
        dialog = NewIntentionDialog(self)

        def update_and_close():
            self.error_label.setText("")
            result = dialog.get_result()
            if not result:
                return
            if result:
                try:
                    self.bayesNet.add_intention(result)
                except AssertionError as e:
                    self.error_label.setText(str(e))
            # update view!
            self.create_fields()
            self.intention_dropdown.setCurrentText(result)
            # Explicit call is neccessary because set seems not to trigger the callback
            self.influencing_context_selected(result)
        ok_button = dialog.findChild(QPushButton, 'ok')
        ok_button.clicked.connect(update_and_close)
        cancel_button = dialog.findChild(QPushButton, "cancel")
        cancel_button.clicked.connect(dialog.reject)
        dialog.exec_()

    def edit_intention(self):
        """
        Edit the name of the currently selected intention
        """
        # open small dialog to create context
        intention = self.intention_dropdown.currentText()
        dialog = NewIntentionDialog(self, intention=intention)

        def update_and_close():
            self.error_label.setText("")
            result = dialog.get_result()
            if not result:
                return
            if result:
                try:
                    self.bayesNet.edit_intention(intention, result)
                except ValueError as e:
                    self.error_label.setText(str(e))
            self.create_fields()
            self.intention_dropdown.setCurrentText(result)
            # Explicit call is necessary because set seems not to trigger the callback
            self.influencing_context_selected(result)
            dialog.accept()

        ok_button = dialog.findChild(QPushButton, "ok")
        ok_button.setDefault(True)
        ok_button.clicked.connect(update_and_close)
        cancel_button = dialog.findChild(QPushButton, "cancel")
        cancel_button.clicked.connect(dialog.reject)
        dialog.exec_()

    def delete_intention(self):
        """
        Delete the currently selected intention
        """
        self.error_label.setText("")
        intention = self.intention_dropdown.currentText()
        try:
            self.bayesNet.del_intention(intention)
        except AssertionError as e:
            self.error_label.setText(str(e))
        self.create_fields()

    def on_clicked_advanced(self):
        """
        Un/folds the advanced section
        """
        if self.advanced_folded:
            # Change text to up arrow
            self.advanced_label.setText("advanced \u25B2")
            self.advanced_folded = False
            self.advanced_hidden_frame.show()
            self.new_combined_influence_button.show()
        else:
            # Change text back to down arrow
            self.advanced_label.setText("advanced \u25BC")
            self.advanced_folded = True
            self.advanced_hidden_frame.hide()
            self.new_combined_influence_button.hide()

    def new_combined_influence(self):
        """
        Callback for the button
        """
        self.error_label.setText("")
        dialog = NewCombinedContextDialog(
            self, intentions=self.bayesNet.config['intentions'])

        def update_and_close():
            result = dialog.get_result()
            try:
                self.bayesNet.add_combined_influence(
                    intention=result['intention'], contexts=result['contexts'], instantiations=result['instantiations'], value=result['value'])
            except ValueError as e:
                self.error_label.setText(str(e))
            self.create_fields()

        ok_button = dialog.findChild(QPushButton, "pushButton_2")
        ok_button.clicked.connect(update_and_close)
        cancel_button = dialog.findChild(QPushButton, "pushButton_3")
        cancel_button.clicked.connect(dialog.reject)
        dialog.exec_()

    def fill_advanced_table(self):
        '''
        Fill the content of the table containing combined influence values
        '''

        font = QFont()
        font.setPointSize(13)
        self.advanced_table.setParent(None)
        self.advanced_table.deleteLater()
        self.advanced_table = QFrame(self.advanced_hidden_frame)
        self.advanced_table.setLayout(QGridLayout())
        self.advanced_hidden_frame.layout().addWidget(self.advanced_table, 0, 0)
        self.advanced_table.layout().addWidget(QLabel('Intention'), 0, 0)
        self.advanced_table.layout().addWidget(QLabel('|'), 0, 1)
        self.advanced_table.layout().addWidget(QLabel('Contexts'), 0, 2)
        self.advanced_table.layout().addWidget(QLabel('|'), 0, 3)
        self.advanced_table.layout().addWidget(QLabel('Influence Value'), 0, 4)
        self.advanced_table.setFont(font)
        row = 1
        for intention, context_influence in self.bayesNet.config['intentions'].items():
            for context in context_influence:
                if isinstance(context, tuple):
                    for j in range(len(list(context_influence[context]))):
                        key = (list(context_influence[context])[j])
                        # For every combined case make a label and a button
                        self.advanced_table.layout().addWidget(
                            QLabel(f'{intention}'), row, 0)
                        self.advanced_table.layout().addWidget(QLabel('|'), row, 1)
                        # build context String
                        context_string = ""
                        for i, _context in enumerate(context):
                            context_string += f'{_context}={str(key[i])}, '
                        context_string = context_string[:-2]
                        self.advanced_table.layout().addWidget(QLabel(context_string), row, 2)
                        self.advanced_table.layout().addWidget(QLabel('|'), row, 3)
                        self.advanced_table.layout().addWidget(
                            QLabel(f'{list(context_influence[context].values())[0]}'), row, 4)
                        remove_button = QPushButton('remove')
                        remove_button.clicked.connect(lambda _, intention=intention, contexts=context, instantiations=list(
                            context_influence[context].keys())[0]: self.remove_combined_influence(intention, contexts, instantiations))
                        self.advanced_table.layout().addWidget(remove_button, row, 5)
                        row += 1

    def remove_combined_influence(self, intention: str, contexts: tuple, instantiations: tuple):
        self.bayesNet.del_combined_influence(
            intention, contexts, instantiations)
        self.fill_advanced_table()

    def setup_layout(self):
        """
        Setting up the layout of the GUI.
        """

        uic.loadUi(Path(Path(__file__).parent, 'configurator.ui'), self)
        self.grid_layout = self.findChild(QGridLayout, 'gridLayout')
        self.grid_layout.setVerticalSpacing(5)

        self.load_button.clicked.connect(self.load)
        self.save_button.clicked.connect(self.save)
        self.decision_threshold_entry.textChanged.connect(
            self.decision_threshold_changed)

        self.error_label = self.findChild(QLabel, 'error_label')
        self.set_error_label_red()
        self.error_label.setText("")

        self.context_label = self.findChild(QLabel, 'intention_label_frame_2')
        self.intention_label_frame = self.findChild(
            QLabel, 'intention_label_frame')
        self.on_label = self.findChild(QLabel, 'label_3')

        self.context_dropdown = self.findChild(QComboBox, 'context_selection')
        self.context_selection = self.context_dropdown

        self.influencing_context_dropdown = self.findChild(
            QComboBox, 'influencing_context_selection')
        self.influencing_context_selection = self.influencing_context_dropdown

        self.intention_dropdown = self.findChild(
            QComboBox, 'intention_selection')

        self.context_instantiations = defaultdict(dict)

        self.context_selected_frame = self.findChild(QFrame, 'frame_2')
        self.influencing_context_frame = self.findChild(QFrame, 'frame')

        self.intention_instantiations = defaultdict(lambda: defaultdict(dict))

        self.new_context_button = self.findChild(
            QPushButton, 'new_context_button')
        self.new_context_button.clicked.connect(self.new_context)

        self.edit_context_button = self.findChild(
            QPushButton, 'edit_context_button')
        self.edit_context_button.clicked.connect(self.edit_context)
        self.delete_context_button = self.findChild(
            QPushButton, 'delete_context_button')
        self.delete_context_button.clicked.connect(self.delete_context)

        self.new_intention_button = self.findChild(
            QPushButton, 'new_intention_button')
        self.new_intention_button.clicked.connect(self.new_intention)
        self.edit_intention_button = self.findChild(
            QPushButton, 'edit_intention_button')
        self.edit_intention_button.clicked.connect(self.edit_intention)
        self.delete_intention_button = self.findChild(
            QPushButton, 'delete_intention_button')

        self.new_combined_influence_button = self.findChild(
            QPushButton, 'advanced_new_button')
        self.new_combined_influence_button.clicked.connect(
            self.new_combined_influence)

        self.advanced_hidden_frame = self.findChild(QFrame, 'frame_3')
        self.advanced_label.setText("advanced \u25BC")
        self.advanced_label.setParent(self.advanced_hidden_frame)
        self.advanced_folded = False
        self.advanced_label.clicked.connect(self.on_clicked_advanced)

        self.advanced_table = QFrame(self.advanced_hidden_frame)
        self.advanced_hidden_frame.setLayout(QGridLayout())
        self.advanced_hidden_frame.layout().addWidget(self.advanced_table, 0, 0)
        self.on_clicked_advanced()

        # add widgets to layout
        self.grid_layout.addWidget(self.context_label, 0, 0)
        self.grid_layout.addWidget(self.context_dropdown, 0, 1)
        self.grid_layout.addWidget(self.edit_context_button, 0, 2)
        self.grid_layout.addWidget(self.delete_context_button, 0, 3)

        self.grid_layout.addWidget(self.context_selected_frame, 1, 1)

        self.grid_layout.addWidget(self.intention_label_frame, 2, 0)
        self.grid_layout.addWidget(self.influencing_context_dropdown, 2, 1)
        self.grid_layout.addWidget(self.on_label, 2, 2)
        self.grid_layout.addWidget(self.intention_dropdown, 2, 3)
        self.grid_layout.addWidget(self.edit_intention_button, 2, 4)
        self.grid_layout.addWidget(self.delete_intention_button, 2, 5)

        self.grid_layout.addWidget(self.context_instantiations_3, 4, 1)
        self.grid_layout.addWidget(self.decision_threshold_entry, 4, 2)

        self.grid_layout.addWidget(self.advanced_label, 5, 1)

        self.grid_layout.addWidget(self.advanced_new_button, 7, 1)

        self.grid_layout.addWidget(self.load_button, 9, 1)
        self.grid_layout.addWidget(self.save_button, 9, 2)

        # Adding the canvas
        layout = QGridLayout()
        self.canvas_frame.setLayout(layout)
        self.canvas_frame.layout().addWidget(self.win, 0, 0)

    def decision_threshold_changed(self, value):
        """
        Callback for change of the decision threshold.
        """
        self.error_label.setText("")
        try:
            self.bayesNet.change_decision_threshold(
                float(value))
        except AssertionError as e:
            self.error_label.setText(f"{e}")
        except ValueError as e:
            self.error_label.setText(f'Decision Threshold must be a number')

    def set_context_dropdown(self, options: list, command: function = None):
        '''
        This sets the options for a context optionMenu with the options and the corresponding command.

        Args:
            options: A list containing the options in the context dropdown
            command: a function which will be triggered when clicked on the dropdown option
        '''
        if not command:
            command = self.context_selected
        self.context_dropdown.clear()

        if options:
            self.context_dropdown.addItems(list(options))
            max_width = max([QFontMetrics(self.context_dropdown.font()).boundingRect(
                option).width() for option in options])
            self.context_dropdown.setMinimumWidth(
                max_width + 25)
            self.context_dropdown.setCurrentIndex(0)
            self.context_dropdown.currentTextChanged.connect(command)
        else:
            self.context_dropdown.addItem('Context')
            self.context_dropdown.currentTextChanged.connect(command)
        command(self.context_dropdown.currentText())

    def draw_graph(self):
        '''
        This draws the graph from the current config.
        '''
        self.graph_item.set_config(self.bayesNet.config)

    def set_influencing_context_dropdown(self, options: list, command: function = None):
        '''
        This sets the options for the influencing context optionMenu with the options and the corresponding command.

        Args:
            options: A list containing the options in the influencing context dropdown
            command: a function which will be triggered when clicked on the dropdown option
        '''
        if not command:
            command = self.influencing_context_selected
        self.influencing_context_dropdown.clear()
        if options:
            self.influencing_context_dropdown.addItems(list(options))
            max_width = max([QFontMetrics(self.influencing_context_dropdown.font(
            )).boundingRect(option).width() for option in options])
            self.influencing_context_dropdown.setMinimumWidth(
                max_width + 25)  # add some padding
            self.influencing_context_dropdown.setCurrentIndex(0)
            self.influencing_context_dropdown.currentTextChanged.connect(
                command)
        else:
            self.influencing_context_dropdown.addItem('Context')
            self.influencing_context_dropdown.currentIndexChanged.connect(
                command)
        command(self.influencing_context_dropdown.currentText())

    def set_intention_dropdown(self, options: list, command: function = None):
        '''
        This sets the options for a intention optionMenu with the options and the corresponding command

        Args:
            options: A list containing the options in the intention dropdown
            command: a function which will be triggered when clicked on the dropdown option
        '''
        if not command:
            command = self.influencing_context_selected
        self.intention_dropdown.clear()

        if options:
            self.intention_dropdown.addItems(list(options))
            max_width = max([QFontMetrics(self.intention_dropdown.font()).boundingRect(
                option).width() for option in options])
            self.intention_dropdown.setMinimumWidth(
                max_width + 25)  # add some padding
            self.intention_dropdown.setCurrentIndex(0)
            self.intention_dropdown.currentTextChanged.connect(command)
        else:
            self.intention_dropdown.addItem('Intention')
            self.intention_dropdown.currentIndexChanged.connect(command)

        command(self.intention_dropdown.currentText())

    def context_selected(self, context: str):
        """
        Callback for click on context in context dropdown.

        Args:
            context: name of the clicked context
        """

        for active_context, instantiations in self.context_instantiations.items():
            for instantiation, widgets in instantiations.items():
                for widget in widgets:
                    try:
                        widget.deleteLater()
                    except AttributeError:
                        pass  # can not destroy StringVars
                    except Exception as e:
                        # TODO: better logging
                        print(f"couldn't destroy: {e}")

        self.context_instantiations = defaultdict(dict)

        if context not in self.bayesNet.config['contexts']:
            return
        config = self.bayesNet.config

        layout = QGridLayout()
        self.context_selected_frame.setLayout(layout)
        layout = self.context_selected_frame.layout()

        layout.setVerticalSpacing(10)
        row_count = layout.rowCount()

        for instantiation, value in config['contexts'][context].items():
            instantiation_str = str(instantiation) if isinstance(
                instantiation, bool) else instantiation
            label = QLabel(instantiation_str, self.context_selected_frame)
            label.setFont(QFont('Times New Roman', 13))
            line_edit = QLineEdit(str(value), self.context_selected_frame)
            line_edit.setFont(QFont('Times New Roman', 13))
            line_edit.textChanged.connect(lambda text, context=context, instantiation=instantiation: self.apriori_values_changed(text, context=context, instantiation=instantiation)
                                          )

            layout.addWidget(label, row_count, 0)
            layout.addWidget(line_edit, row_count, 1)

            self.context_instantiations[context][instantiation] = (
                label,
                line_edit,
            )
            row_count += 1

    def influencing_context_selected(self, context_or_intention: str):
        """
        Callback for click on context in influencing context dropdown.

        Args:
            context_or_intention: name of the clicked context
        """
        for active_intention, active_context in self.intention_instantiations.items():
            for active_context, instantiations in active_context.items():
                for instantiation, widgets in instantiations.items():
                    for widget in widgets:
                        try:
                            widget.deleteLater()
                        except AttributeError:
                            pass  # can not destroy StringVars
                        except Exception as e:
                            # TODO: better logging
                            print(f"couldn't destroy: {e}")
        intention = self.intention_dropdown.currentText()
        context = self.influencing_context_selection.currentText()
        if context not in self.bayesNet.config['contexts'] or intention not in self.bayesNet.config['intentions']:
            return

        self.intention_instantiations = defaultdict(lambda: defaultdict(dict))

        layout = QGridLayout()
        self.influencing_context_frame.setLayout(layout)
        layout = self.influencing_context_frame.layout()

        layout.setVerticalSpacing(10)
        row_count = layout.rowCount()

        for instantiation, value in self.bayesNet.config['intentions'][intention][context].items():
            instantiation_label = QLabel(
                f'Influence of {context}:{instantiation} on {intention}: LOW', self.influencing_context_frame)
            instantiation_label.setFont(QFont('Times New Roman', 13))
            slider = QSlider(Qt.Horizontal, self.influencing_context_frame)
            slider.setFixedSize(100, 20)
            high_label = QLabel('HIGH', self.influencing_context_frame)
            high_label.setFont(QFont('Times New Roman', 13))

            layout.addWidget(instantiation_label, row_count, 0)
            layout.addWidget(slider, row_count, 1)
            layout.addWidget(high_label, row_count, 2)

            slider.setMinimum(0)
            slider.setMaximum(5)
            slider.setTickInterval(1)
            slider.setValue(value)
            slider.valueChanged.connect(lambda value, context=context, intention=intention,
                                        instantiation=instantiation: self.influence_values_changed(value, context, intention, instantiation))

            self.intention_instantiations[intention][context][instantiation] = (
                instantiation_label,
                slider,
                high_label,
            )

            row_count += 1

    def load(self):
        """
        opens a askopenfilename dialog to load a config
        """
        options = QFileDialog.Options()
        options |= QFileDialog.ReadOnly
        fileName, _ = QFileDialog.getOpenFileName(
            None, "Choose Config", "", "Yaml files (*.yml);;All Files (*)", options=options)
        if fileName:
            try:
                self.error_label.setText("loading BayesNet...")
                self.bayesNet.load(fileName)
                self.error_label.setText("")
            except AssertionError as e:
                self.error_label.setText(str(e))
            except Exception as e:
                self.error_label.setText(str(e))
        self.create_fields()

    def save(self):
        """
        opens a asksaveasfilename dialog to save a config
        """
        filetypes = "Yaml files (*.yml);;All Files (*)"
        save_filepath, _ = QFileDialog.getSaveFileName(
            None, "Save Config", "", filetypes)
        if save_filepath:
            self.bayesNet.save(save_filepath)

    def apriori_values_changed(self, *args, context, instantiation):
        """
        Callback for change of the apriori values.

        Args:
            context: name of the context
            instantiation: name of the corresponding instantiation
        """
        # update config
        self.error_label.setText("")
        try:
            self.bayesNet.change_context_apriori_value(context=context, instantiation=instantiation, value=float(
                self.context_instantiations[context][instantiation][1].text()))
        except AssertionError as e:
            self.error_label.setText(str(e))
        except ValueError as e:
            self.error_label.setText(
                f'Apriori probability of context "{context}.{instantiation}" is not a number')

    def influence_values_changed(self, value, context, intention, instantiation):
        """
        Callback for change of the influence values.

        Args:
            value: new influence value
            context: name of the context
            intention: name of the intention
            instantiation: name of the corresponding instantiation
        """
        self.error_label.setText("")
        try:
            self.bayesNet.change_influence_value(
                intention=intention, context=context, instantiation=instantiation, value=int(value))
        except AssertionError as e:
            self.error_label.setText(str(e))

        return value


class TwoLayerGraph(pg.GraphItem):
    """
    Graph Visualization for the two layer bayesian network
    """

    def __init__(self, dist=10, size=3, line_width=[0, 25], pxMode=False, **kwds):
        super().__init__(**kwds)
        self.dist = dist
        self.size = size
        self.line_width = line_width
        self.pxMode = pxMode
        self.unfolded_context = set()
        self.textItems = []

    def _set_pos(self):
        """
        Add all the data points in the pos array
        """
        i = 0
        added_context = set()
        for intention, context_dict in self.config['intentions'].items():
            position = (self.dist, self.dist/2 + i*self.dist)
            self.data["pos"].append(position)
            self.data["names"].append(intention)
            self.data["intention_indices"].append(i)
            # self.data["mapping"][position] = intention
            i += 1
            for context, instantiation_dict in context_dict.items():
                if context not in added_context and isinstance(context, str):
                    if context in self.unfolded_context:
                        for instatiation in instantiation_dict:
                            if f"{context}:{instatiation}" not in added_context and isinstance(context, str):
                                position = (0, i*self.dist)
                                self.data["pos"].append(position)
                                self.data["names"].append(
                                    (context, instatiation))
                                # f"{context}:{instatiation}")
                                # TODO: use tuple here and check for type on usage
                                # TODO: this may be 'instantiation_indices'
                                self.data["instantiation_indices"].append(i)
                                # self.data["context_indices"].append(i)
                                # self.data["mapping"][position] = f"{context}:{instatiation}"
                                added_context.add(f"{context}:{instatiation}")
                                i += 1
                    else:
                        position = (0, i*self.dist)
                        self.data["pos"].append(position)
                        self.data["names"].append(context)
                        self.data["context_indices"].append(i)
                        # self.data["mapping"][position] = context
                        added_context.add(context)
                        i += 1

    def _set_adj(self):
        """
        Add all the connections in the adj array
        """
        # TODO: merge context and instantiations indices
        left_side = self.data["context_indices"] + \
            self.data["instantiation_indices"]
        self.data["adj"] = list(itertools.product(
            left_side, self.data["intention_indices"]))

    def _set_pen(self):
        """
        Add all the pens for the connections in the pen array 
        """
        for start, end in self.data["adj"]:
            if start in self.data["context_indices"] or start in self.data["instantiation_indices"]:
                context = self.data["names"][start]
                intention = self.data["names"][end]
            else:
                context = self.data["names"][end]
                intention = self.data["names"][start]
            color = pg.mkPen().color()
            normalized_mean = np.mean(
                list(self.config["intentions"][intention][context].values()))/5.0

            if start in self.data["instantiation_indices"]:
                # HAck: TODO: this is problematic if the context has a colon in name
                context, instantiation = self.data["names"][start]
                # TODO. problem with default dict or problem with type - for "human holding object" bool is used and it does not work...
                normalized_mean = self.config["intentions"][intention][context][instantiation] / 5.0
            alpha = color.alpha()
            # TODO: Maybe color can even be scaled with the normalized mean
            red = color.red() if normalized_mean else 255
            green = color.green() if normalized_mean else 0
            blue = color.blue() if normalized_mean else 0
            width = self.line_width[0] + \
                (self.line_width[1] - self.line_width[0]) * normalized_mean

            self.data["pen"].append(np.array([(red, green, blue, alpha, width)], dtype=[
                ('red', np.uint8), ('green', np.uint8), ('blue', np.uint8), ('alpha', np.uint8), ('width', np.uint8)]))

    def _set_text(self):
        """
        Place all the texts for Context and Intention Names
        """
        for i in self.textItems:
            i.scene().removeItem(i)
        self.textItems = []
        # self.data["mapping"].items():
        for position, label in zip(self.data["pos"], self.data["names"]):
            # TODO: change color
            if isinstance(label, tuple):
                label = f"{label[0]}:{label[1]}"
            text_item = pg.TextItem(label, anchor=(0.5, 0.5))
            text_item.setParentItem(self)
            text_item.setPos(*position)
            self.textItems.append(text_item)

    def set_config(self, config):
        """
        Uses the config to set the data
        """
        # extract every needed field for setData from config

        self.config = config
        self.data = {"mapping": {}, "pos": [],
                     "adj": [], "pen": [], "names": [], "context_indices": [], "intention_indices": [], "instantiation_indices": []}
        self._set_pos()
        self._set_adj()
        self._set_pen()
        self._set_text()

        self.setData(pos=np.array(self.data["pos"]), adj=np.array(
            self.data["adj"]), pen=np.array(self.data["pen"]), size=self.size, pxMode=self.pxMode)

    def mousePressEvent(self, event):
        """
        Handler for the mouse click event
        """
        click_pos = event.pos()
        # Context
        if click_pos.x() > 0 - (self.size/2.0) and click_pos.x() < 0 + (self.size/2.0):
            # self.data["mapping"].items():
            i = 0
            for position, name in zip(self.data["pos"], self.data["names"]):
                if click_pos.y() > position[1] - (self.size/2.0) and click_pos.y() < position[1] + (self.size/2.0) and position[0] == 0:
                    # click on folded context
                    if i in self.data["context_indices"]:
                        self.unfolded_context.add(name)
                    # click on one of the instantiations of an unfolded context
                    if i in self.data["instantiation_indices"]:
                        # HAck: TODO: this is problematic if the context has a colon in name
                        context, instantiation = self.data["names"][i]
                        self.unfolded_context.remove(context)
                i += 1
        # Intention
        if click_pos.x() > self.dist - (self.size/2.0) and click_pos.x() < self.dist + (self.size/2.0):
            pass
            # TODO: maybe set the focus to the corresponding field

        self.set_config(self.config)



'''
Tests for adding intention variables
'''

# System imports
from copy import deepcopy
import pytest
# 3rd party imports

# local imports
from tests import N
from CoBaIR.bayes_net import BayesNet, default_to_regular, load_config

# end file header
__author__ = 'Adrian Lubitz'



def test_edit_context_from_empty():
    """
    Test editing on a non existing config
    """
    bn = BayesNet()
    old_config1 = deepcopy(bn.config)
    # I assume this will throw an Error!
    with pytest.raises(ValueError):
        bn.edit_context('the context', {
            'inst_a': 0.3, 'inst_b': 0.3, 'inst_c': 0.4})
    assert old_config1 == bn.config

def test_edit_context_from_existing_new_name():
    """
    Test editing on an existing config - only changing the name of the context
    """
    bn = BayesNet()
    bn.load('small_example.yml')
    new_name = 'commandos de speechos'
    old_context = 'speech commands'
    old_config = deepcopy(bn.config)
    # I assume this will throw an Error!
    instantiations = deepcopy(bn.config['contexts'][old_context])
    bn.edit_context(old_context, instantiations, new_name=new_name)
    assert new_name in bn.config['contexts']
    assert old_context not in bn.config['contexts']
    assert instantiations == bn.config['contexts'][new_name]
    for intention in bn.config['intentions']:
        assert new_name in bn.config['intentions'][intention]
        assert old_context not in bn.config['intentions'][intention]
        assert old_config['intentions'][intention][old_context] == bn.config['intentions'][intention][new_name]

def test_edit_context_from_existing_new_instantiations():
    """
    Test editing on an existing config - changing the names and values of the instantiations
    """
    bn = BayesNet()
    bn.load('small_example.yml')
    context = 'speech commands'
    old_config = deepcopy(bn.config)
    # I assume this will throw an Error!
    instantiations = {
        'inst_a': 0.3, 'inst_b': 0.3, 'inst_c': 0.4}
    bn.edit_context(context, instantiations)

    assert context in bn.config['contexts']
    assert instantiations == bn.config['contexts'][context]
    for intention in bn.config['intentions']:
        assert context in bn.config['intentions'][intention]
        # all influence values will be set to zero because I don't know if the instantiation is a total new or just corrected a typo
        for instantiation in instantiations:
            assert instantiation in bn.config['intentions'][intention][context]
            assert 0 == bn.config['intentions'][intention][context][instantiation]

def test_edit_context_from_existing_adjust_instantiations():
    """
    Test editing on an existing config - changing one name of the instantiations
    """
    bn = BayesNet()
    bn.load('small_example.yml')
    context = 'speech commands'
    old_config = deepcopy(bn.config)
    old_instantiation_name = 'pickup'
    new_instantiation_name = 'pick up'
    # I assume this will throw an Error!
    instantiations = deepcopy(bn.config['contexts'][context])
    old_value = instantiations[old_instantiation_name]
    del(instantiations[old_instantiation_name])
    instantiations[new_instantiation_name] = old_value
    bn.edit_context(context, instantiations)

    assert context in bn.config['contexts']
    assert instantiations == bn.config['contexts'][context]
    for intention in bn.config['intentions']:
        assert context in bn.config['intentions'][intention]
        # all influence values will be set to zero because I don't know if the instantiation is a total new or just corrected a typo
        for instantiation in instantiations:
            assert instantiation in bn.config['intentions'][intention][context]
            if instantiation != new_instantiation_name:
                assert old_config['intentions'][intention][context][instantiation] ==bn.config['intentions'][intention][context][instantiation]
            else:
                assert bn.config['intentions'][intention][context][instantiation] == 0

def test_edit_context_from_existing_new_instantiations_and_new_name():
    """
    Test editing on an existing config - changing the names and values of the instantiations and the context name
    """
    bn = BayesNet()
    bn.load('small_example.yml')
    new_name = 'commandos de speechos'
    old_context = 'speech commands'
    old_config = deepcopy(bn.config)
    # I assume this will throw an Error!
    instantiations = {
        'inst_a': 0.3, 'inst_b': 0.3, 'inst_c': 0.4}
    bn.edit_context(old_context, instantiations, new_name=new_name)
    assert new_name in bn.config['contexts']
    assert instantiations == bn.config['contexts'][new_name]

    for intention in bn.config['intentions']:
        assert new_name in bn.config['intentions'][intention]
        # all influence values will be set to zero because I don't know if the instantiation is a total new or just corrected a typo
        for instantiation in instantiations:
            assert instantiation in bn.config['intentions'][intention][new_name]
            assert 0 == bn.config['intentions'][intention][new_name][instantiation]

def test_edit_context_from_existing_adjust_instantiations():
    """
    Test editing on an existing config - changing one name of the instantiations and the context name
    """
    bn = BayesNet()
    bn.load('small_example.yml')
    new_name = 'commandos de speechos'
    old_context = 'speech commands'
    old_config = deepcopy(bn.config)
    old_instantiation_name = 'pickup'
    new_instantiation_name = 'pick up'
    # I assume this will throw an Error!
    instantiations = deepcopy(bn.config['contexts'][old_context])
    old_value = instantiations[old_instantiation_name]
    del(instantiations[old_instantiation_name])
    instantiations[new_instantiation_name] = old_value
    bn.edit_context(old_context, instantiations, new_name=new_name)

    assert new_name in bn.config['contexts']
    assert instantiations == bn.config['contexts'][new_name]
    for intention in bn.config['intentions']:
        assert new_name in bn.config['intentions'][intention]
        # all influence values will be set to zero because I don't know if the instantiation is a total new or just corrected a typo
        for instantiation in instantiations:
            assert instantiation in bn.config['intentions'][intention][new_name]
            if instantiation != new_instantiation_name:
                assert old_config['intentions'][intention][old_context][instantiation] == bn.config['intentions'][intention][new_name][instantiation]
            else:
                assert bn.config['intentions'][intention][new_name][instantiation] == 0




'''
Tests for remove combined context influence
'''

# System imports
from copy import deepcopy
import pytest

from collections import defaultdict

# local imports
from tests import N
from CoBaIR.bayes_net import BayesNet, default_to_regular, load_config

# end file header
__author__ = 'Arunima Gopikrishnan'

config = {'intentions': defaultdict(lambda: defaultdict(
        lambda: defaultdict(int))), 'contexts': defaultdict(lambda: defaultdict(float))}
def test_remove_combined_influence_from_empty():
    """
    Test removing combined context influence on a non existing config
    """
    bn = BayesNet()
    contexts = 'hand over tool'
    intention = ('speech commands','human activity')
    instantiations = ('pickup','working')
    # I assume this will throw an Error!
    with pytest.raises(ValueError):
        bn.del_combined_influence(intention, contexts, instantiations)



def test_remove_existing_add_new_combined_context_influence():
    """
    Test removing combined context influence from the loaded config's existing combined context influence
    and adding combined context influence into the loaded config's existing combined context influence
    """
    
    bn = BayesNet()
    bn.load('small_example.yml')
    original_config = deepcopy(bn.config)
 #   for intention, context_influence in bn.config['intentions'].items():
#            bn._create_combined_context(context_influence)
    intention = 'pick up tool'
    contexts = ('speech commands','human activity')
    instantiations = ('pickup','working')
    value = 5
    bn.del_combined_influence(intention, contexts, instantiations)
    altered_config = deepcopy(bn.config)
    assert altered_config != original_config
    bn.add_combined_influence(intention, contexts, instantiations,value)
    assert original_config == bn.config
    



from setuptools import setup
from setuptools import find_packages
import CoBaIR
import os

with open('requirements/requirements.txt') as f:
    requirements = f.read().splitlines()

# TODO: if TAG in pipeline given use tag
if 'CI_COMMIT_TAG' in os.environ:
    VERSION = os.environ['CI_COMMIT_TAG']
else:
    VERSION = '0.0.0'


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


setup(name='CoBaIR',
      version=VERSION,
      description='CoBaIR is a python lib for Context Based Intention Recognition',
      author='Adrian Lubitz',
      author_email='Adrian.Lubitz@dfki.de',
      license='BSD 3-Clause',
      keywords="intention recognition context human machine interaction robot bayesian",
      url="https://github.com/dfki-ric/CoBaIR",
      install_requires=requirements,
      packages=find_packages(),
      long_description=read('README.md'),
      long_description_content_type="text/markdown",
      classifiers=[
          "Development Status :: 3 - Alpha",
          "Environment :: X11 Applications :: Qt",
          "Intended Audience :: Science/Research",
          "Programming Language :: Python :: 3.8",
          "Topic :: Scientific/Engineering :: Artificial Intelligence",
          "Topic :: Utilities",
          "License :: OSI Approved :: BSD License",
      ],
      python_requires='>=3.8')



'''
Tests for adding intention variables
'''

# System imports
from copy import deepcopy
import pytest

# 3rd party imports

# local imports
from tests import N
from CoBaIR.bayes_net import BayesNet, default_to_regular, load_config

# end file header
__author__ = 'Adrian Lubitz'
bn = BayesNet()
bn.load('small_example.yml')


def test_infer_no_evidence():
    """
    Test inference without evidence
    """
    probabilities = {'pick up tool': 0.602803738317757,
                     'hand over tool': 0.397196261682243}

    max_intention, inference = bn.infer({})
    for intention, probability in inference.items():
        assert round(abs(probability-probabilities[intention]), 7) == 0


def test_infer_unrelated_evidence():
    """
    Test inference with unrelated evidence of different types
    """
    probabilities = {'pick up tool': 0.602803738317757,
                     'hand over tool': 0.397196261682243}

    max_intention, inference = bn.infer({'some context': 'is not important', 'another context': '',
                                         'unhashable context': {}, 'int context': 1, 'obj context': bn})
    for intention, probability in inference.items():
        assert round(abs(probability-probabilities[intention]), 7) == 0


def test_infer_single_evidence():
    """
    Test inference with single evidence context present
    """
    probabilities = {'pick up tool': 0.7821782178217822,
                     'hand over tool': 0.21782178217821785}

    max_intention, inference = bn.infer(
        {'speech commands': 'pickup', 'unhashable context': {}})
    for intention, probability in inference.items():
        assert round(abs(probability-probabilities[intention]), 7) == 0


def test_infer_multiple_evidence():
    """
    Test inference with multiple evidence context present
    """
    probabilities = {'hand over tool': 0.3797468354430379,
                     'pick up tool': 0.620253164556962}

    max_intention, inference = bn.infer(
        {'speech commands': 'pickup', 'human holding object': False, 'human activity': 'idle', 'unhashable context': {}})
    for intention, probability in inference.items():
        assert round(abs(probability-probabilities[intention]), 7) == 0


def test_infer_combined_evidence():
    """
    Test inference for the combined case
    """
    probabilities = {'hand over tool': 0.26229508196721313,
                     'pick up tool': 0.7377049180327869}

    max_intention, inference = bn.infer(
        {'speech commands': 'pickup', 'human holding object': True, 'human activity': 'idle', 'unhashable context': {}})
    for intention, probability in inference.items():
        assert round(abs(probability-probabilities[intention]), 7) == 0


def test_infer_overlapping_combined_evidence():
    """
    Test inference for combined case if two combinations are overlapping - not implemented yet!
    """
    # probabilities = {'hand over tool': 0.26229508196721313,
    #                  'pick up tool': 0.7377049180327869}

    # inference = self.bn.infer(
    #     {'speech commands': 'pickup', 'human holding object': True, 'human activity': 'idle', 'unhashable context': {}})
    # print(inference)
    # for intention, probability in inference.items():
    #     self.assertAlmostEqual(probability, probabilities[intention])
    pass


def test_infer_invalid_evidence():
    """
    Test inference
    """
    probabilities = {'pick up tool': 0.7821782178217822,
                     'hand over tool': 0.21782178217821785}

    with pytest.raises(ValueError):
        bn.infer(
            {'speech commands': 'ERROR', 'unhashable context': {}})



'''
Tests for Loading yaml format config file
'''

# System imports
import pytest
from collections import defaultdict
# 3rd party imports
from yaml.parser import ParserError

# local imports
from tests import N
from CoBaIR.bayes_net import BayesNet, load_config

# end file header
__author__ = 'Arunima Gopikrishnan'

# TODO: add more negative cases


def test_loading_invalid_file_format():
    """
    Test loading invalid file format.
    """

    # I assume this will throw an Error!
    with pytest.raises(ParserError):
        load_config('README.md')


def test_loading_non_existing_file():
    """
    Test loading non existing file.
    """

    # I assume this will throw an Error!
    with pytest.raises(FileNotFoundError):
        load_config('EATME.md')


def test_loading_invalid_file_name():
    """
    Test loading invalid file name - try to loading non-existing file name
    """

    # I assume this will throw an Error!
    with pytest.raises(FileNotFoundError):
        load_config('small.yml')


def test_loading_invalid_yml_file():
    """
    Test loading invalid file name
    """

    config = load_config('small_example_invalid.yml')

    # I assume this will throw an Error!- AssertionError: Influence Value for pick up tool.speech commands.handover must be an integer between 0 and 5! Is 10
    with pytest.raises(AssertionError):
        BayesNet(config)
        assert BayesNet(config).valid == False


def test_loading_valid_yml_file():
    """
    Test loading valid file name - validata_config() function will validate that the current config follows the correct format.
    """

    config = load_config('small_example.yml')
    bn = BayesNet(config)
    # valid config
    assert bn.valid == True



'''
Tests for adding intention variables
'''

# System imports
import pytest
from collections import defaultdict
# 3rd party imports

# local imports
from tests import N
from CoBaIR.bayes_net import BayesNet, config_to_default_dict, default_to_regular, load_config

# end file header
__author__ = 'Adrian Lubitz'

# TODO: add more negative cases


def test_add_to_empty():
    """
    Test adding an intention to an empty config
    """
    config = {'intentions': defaultdict(lambda: defaultdict(
        lambda: defaultdict(int))), 'contexts': defaultdict(lambda: defaultdict(float))}
    intention = 'some intention'

    config['intentions'][intention] = {}
    bn = BayesNet()
    # I assume this will throw an Error!
    with pytest.raises(AssertionError):
        bn.add_intention('some intention')
    # Making sure tmp_config will be maintained
    assert config_to_default_dict(bn.config) == config_to_default_dict(config)


def test_add_n_to_empty(n=N):
    """
    Test adding n intentions to an empty config
    """
    config = {'intentions': defaultdict(lambda: defaultdict(
        lambda: defaultdict(int))), 'contexts': defaultdict(lambda: defaultdict(float))}
    intention = 'intention'
    bn = BayesNet()
    # I assume this will throw an Error!
    for i in range(n):
        config['intentions'][f'{intention}_{i}'] = {}
        with pytest.raises(AssertionError):
            bn.add_intention(f'{intention}_{i}')
    # Making sure tmp_config will be maintained
    assert config_to_default_dict(bn.config) == config_to_default_dict(config)


def test_add_to_existing():
    """
    Test adding an intention to an existing config
    """
    config = load_config('small_example.yml')
    bn = BayesNet(config)
    bn.add_intention('some intention')


def test_add_n_to_existing(n=N):
    """
    Test adding n intentions to an existing config
    """
    config = load_config('small_example.yml')
    bn = BayesNet(config)
    for i in range(n):
        bn.add_intention(f'intention_{i}')


def test_add_to_existing_loaded():
    """
    Test adding an intention to an existing config
    """
    bn = BayesNet()
    bn.load('small_example.yml')
    bn.add_intention('some intention')


def test_add_n_to_existing_loaded(n=N):
    """
    Test adding n intentions to an existing config
    """
    bn = BayesNet()
    bn.load('small_example.yml')
    for i in range(n):
        bn.add_intention(f'intention_{i}')
